# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.groups
r"""
Constellations

A constellation is a tuple `(g_0, g_1, \dots, g_k)` of permutations
such that the product `g_0 g_1 ... g_k` is the identity. One often
assumes that the group generated by `g_0, g_1, \dots, g_k` acts
transitively ([LZ2004]_ definition 1). Geometrically, it corresponds
to a covering of the 2-sphere ramified over `k` points (the transitivity
condition corresponds to the connectivity of the covering).

EXAMPLES::

    sage: c = Constellation(['(1,2)', '(1,3)', None])
    sage: c
    Constellation of length 3 and degree 3
    g0 (1,2)(3)
    g1 (1,3)(2)
    g2 (1,3,2)
    sage: C = Constellations(3,4); C
    Connected constellations of length 3 and degree 4 on {1, 2, 3, 4}
    sage: C.cardinality()  # long time
    426

    sage: C = Constellations(3, 4, domain=('a', 'b', 'c', 'd'))
    sage: C
    Connected constellations of length 3 and degree 4 on {'a', 'b', 'c', 'd'}
    sage: c = C(('a','c'),(('b','c'),('a','d')), None)
    sage: c
    Constellation of length 3 and degree 4
    g0 ('a','c')('b')('d')
    g1 ('a','d')('b','c')
    g2 ('a','d','c','b')
    sage: c.is_connected()
    True
    sage: c.euler_characteristic()
    2
    sage: TestSuite(C).run()
"""

# ****************************************************************************
#       Copyright (C) 2015-2016 Vincent Delecroix <20100.delecroix@gmail.com>
#                               Frédéric Chapoton <fchapoton2@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from itertools import repeat, product

from sage.structure.element import parent
from sage.structure.parent import Parent
from sage.structure.element import Element
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.richcmp import (op_NE, op_EQ, richcmp_not_equal,
                                    rich_to_bool)

from sage.rings.integer import Integer
from sage.combinat.partition import Partition
from sage.misc.misc_c import prod
from sage.misc.lazy_import import lazy_import
from sage.categories.groups import Groups

lazy_import('sage.graphs.graph', 'Graph')
lazy_import('sage.graphs.digraph', 'DiGraph')
lazy_import('sage.groups.perm_gps.permgroup_named', 'SymmetricGroup')

# constructors


def Constellations(*data, **options):
    r"""
    Build a set of constellations.

    INPUT:

    - ``profile`` -- an optional profile

    - ``length`` -- an optional length

    - ``degree`` -- an optional degree

    - ``connected`` -- an optional boolean

    EXAMPLES::

        sage: Constellations(4,2)
        Connected constellations of length 4 and degree 2 on {1, 2}

        sage: Constellations([[3,2,1],[3,3],[3,3]])
        Connected constellations with profile ([3, 2, 1], [3, 3], [3, 3]) on {1, 2, 3, 4, 5, 6}
    """
    profile = options.get('profile', None)
    length = options.get('length', None)
    degree = options.get('degree', None)
    connected = options.get('connected', True)
    domain = options.get('domain', None)
    if domain is not None:
        domain = tuple(domain)

    if data:
        if len(data) == 1:
            if isinstance(data[0], (tuple, list)):
                profile = data[0]
            else:
                length = Integer(data[0])
        elif len(data) == 2:
            length = Integer(data[0])
            degree = Integer(data[1])

    if profile:
        profile = tuple(map(Partition, profile))
        return Constellations_p(profile, domain, bool(connected))
    elif degree is not None and length is not None:
        if domain is None:
            sym = SymmetricGroup(degree)
        else:
            sym = SymmetricGroup(domain)
            if len(sym.domain()) != degree:
                raise ValueError("the size of the domain should be equal to the degree")

        return Constellations_ld(Integer(length), Integer(degree),
                                 sym, bool(connected))
    else:
        raise ValueError("you must either provide a profile or a pair (length, degree)")


def Constellation(g=None, mutable=False, connected=True, check=True):
    r"""
    Constellation.

    INPUT:

    - ``g`` -- list of permutations

    - ``mutable`` -- boolean (default: ``False``); whether the result is
      mutable or not

    - ``connected`` -- boolean (default: ``True``); whether the result should
      be connected

    - ``check`` -- boolean (default: ``True``); whether or not to check. If it
      is ``True``, then the list ``g`` must contain no ``None``.

    EXAMPLES:

    Simple initialization::

        sage: Constellation(['(0,1)','(0,3)(1,2)','(0,3,1,2)'])
        Constellation of length 3 and degree 4
        g0 (0,1)(2)(3)
        g1 (0,3)(1,2)
        g2 (0,3,1,2)

    One of the permutation can be omitted::

        sage: Constellation(['(0,1)', None, '(0,4)(1,2,3)'])
        Constellation of length 3 and degree 5
        g0 (0,1)(2)(3)(4)
        g1 (0,3,2,1,4)
        g2 (0,4)(1,2,3)

    One can define mutable constellations::

        sage: Constellation(([0,2,1], [2,1,0], [1,2,0]), mutable=True)
        Constellation of length 3 and degree 3
        g0 (0)(1,2)
        g1 (0,2)(1)
        g2 (0,1,2)
    """
    l = len(g)
    sym, _ = perms_sym_init([x for x in g if x is not None])
    d = len(sym.domain())
    return Constellations(l, d,
                          domain=sym.domain(),
                          connected=connected)(g, mutable=mutable, check=check)

# classes


class Constellation_class(Element):
    r"""
    Constellation.

    A constellation or a tuple of permutations `(g_0,g_1,...,g_k)`
    such that the product `g_0 g_1 ... g_k` is the identity.
    """

    def __init__(self, parent, g, connected, mutable, check):
        r"""
        TESTS::

            sage: c = Constellation([[1,2,0],[0,2,1],[1,0,2],None])
            sage: c == loads(dumps(c))
            True
            sage: g0 = '(0,1)(2,4)'
            sage: g1 = '(0,3)(1,4)'
            sage: g2 = '(2,4,3)'
            sage: g3 = '(0,3)(1,2)'
            sage: c0 = Constellation([g0,g1,g2,g3])
            sage: c0 == Constellation([None,g1,g2,g3])
            True
            sage: c0 == Constellation([g0,None,g2,g3])
            True
            sage: c0 == Constellation([g0,g1,None,g3])
            True
            sage: c0 == Constellation([g0,g1,g2,None])
            True
        """
        Element.__init__(self, parent)
        self._connected = connected
        self._mutable = mutable
        self._g = g
        if check:
            self._check()

    def __hash__(self):
        r"""
        Return a hash for ``self``.

        EXAMPLES::

            sage: c = Constellation(([0,2,1],[2,1,0],[1,2,0]), mutable=False)
            sage: hash(c) == hash(tuple(c._g))
            True
        """
        if self._mutable:
            raise ValueError("cannot hash mutable constellation")
        return hash(tuple(self._g))

    def set_immutable(self):
        r"""
        Do nothing, as ``self`` is already immutable.

        EXAMPLES::

            sage: c = Constellation(([0,2,1],[2,1,0],[1,2,0]), mutable=False)
            sage: c.set_immutable()
            sage: c.is_mutable()
            False
        """
        self._mutable = False

    def is_mutable(self):
        r"""
        Return ``False`` as ``self`` is immutable.

        EXAMPLES::

            sage: c = Constellation(([0,2,1],[2,1,0],[1,2,0]), mutable=False)
            sage: c.is_mutable()
            False
        """
        return self._mutable

    def switch(self, i, j0, j1):
        r"""
        Perform the multiplication by the transposition `(j0, j1)` between the
        permutations `g_i` and `g_{i+1}`.

        The modification is local in the sense that it modifies `g_i`
        and `g_{i+1}` but does not modify the product `g_i g_{i+1}`. The new
        constellation is

        .. MATH::

            (g_0, \ldots, g_{i-1}, g_{i} (j0 j1), (j0 j1) g_{i+1}, g_{i+2}, \ldots, g_k)

        EXAMPLES::

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None], mutable=True); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2)(3)
            g2 (0,1,3,2,4)
            sage: c.is_mutable()
            True
            sage: c.switch(1,2,3); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2,3)
            g2 (0,1,3,4)(2)
            sage: c._check()
            sage: c.switch(2,1,3); c
            Constellation of length 3 and degree 5
            g0 (0,1,4,2,3)
            g1 (0)(1,4)(2,3)
            g2 (0,3,4)(1)(2)
            sage: c._check()
            sage: c.switch(0,0,1); c
            Constellation of length 3 and degree 5
            g0 (0)(1,4,2,3)
            g1 (0,4,1)(2,3)
            g2 (0,3,4)(1)(2)
            sage: c._check()
        """
        if not self._mutable:
            raise ValueError("this constellation is immutable."
                             " Take a mutable copy first.")
        S = SymmetricGroup(list(range(self.degree())))
        tr = S((j0, j1))
        i = int(i)
        if i < 0 or i >= len(self._g):
            raise ValueError("index out of range")

        ii = i + 1
        if ii == len(self._g):
            ii = 0
        self._g[i] = self._g[i] * tr
        self._g[ii] = tr * self._g[ii]

    def euler_characteristic(self):
        r"""
        Return the Euler characteristic of the surface.

        ALGORITHM:

        Hurwitz formula

        EXAMPLES::

            sage: c = Constellation(['(0,1)', '(0,2)', None])
            sage: c.euler_characteristic()
            2

            sage: c = Constellation(['(0,1,2,3)','(1,3,0,2)', '(0,3,1,2)', None])
            sage: c.euler_characteristic()
            -4

        TESTS::

            sage: parent(c.euler_characteristic())
            Integer Ring
        """
        return Integer(self.degree() * 2 -
                       sum(sum(j - 1 for j in self.profile(i))
                           for i in range(self.length())))

    def genus(self):
        r"""
        Return the genus of the surface.

        EXAMPLES::

            sage: c = Constellation(['(0,1)', '(0,2)', None])
            sage: c.genus()
            0

            sage: c = Constellation(['(0,1)(2,3,4)','(1,3,4)(2,0)', None])
            sage: c.genus()
            1

        TESTS::

            sage: parent(c.genus())
            Integer Ring
        """
        return 1 - self.euler_characteristic() // 2

    def _check(self):
        r"""
        Check that the constellation is valid and if not raise
        :exc:`ValueError`.

        TESTS::

            sage: c = Constellation([[0,1],[1,0]], mutable=True, check=False)
            sage: c._check()
            Traceback (most recent call last):
            ...
            ValueError: the product is not identity

            sage: c = Constellation([[0,1],[0,1]], mutable=True, check=False)
            sage: c._check()
            Traceback (most recent call last):
            ...
            ValueError: not connected
        """
        d = self.degree()
        Sd = self.parent()._sym

        if prod(self._g, Sd.one()) != Sd.one():
            raise ValueError("the product is not identity")

        if self._connected and not perms_are_connected(self._g, d):
            raise ValueError("not connected")

    def __copy__(self):
        r"""
        Return a copy of ``self``.

        TESTS::

            sage: c = Constellation([[0,2,1],[1,0,2],[2,1,0],None])
            sage: c == copy(c)
            True
            sage: c is copy(c)
            False
            sage: c = Constellation([[0,2,1],[1,0,2],[2,1,0],None],mutable=True)
            sage: c == copy(c)
            True
            sage: c is copy(c)
            False
        """
        return self.parent()(list(self._g),
                             check=False,
                             mutable=self._mutable)

    copy = __copy__

    def mutable_copy(self):
        r"""
        Return a mutable copy of ``self``.

        EXAMPLES::

            sage: c = Constellation(([0,2,1],[2,1,0],[1,2,0]), mutable=False)
            sage: d = c.mutable_copy()
            sage: d.is_mutable()
            True
        """
        return self.parent()(list(self._g),
                             check=False,
                             mutable=True)

    # GENERAL PROPERTIES

    def is_connected(self):
        r"""
        Test of connectedness.

        EXAMPLES::

            sage: c = Constellation(['(0,1)(2)', None, '(0,1)(2)'], connected=False)
            sage: c.is_connected()
            False
            sage: c = Constellation(['(0,1,2)', None], connected=False)
            sage: c.is_connected()
            True
        """
        if self._connected:
            return True
        else:
            return perms_are_connected(self._g, self.degree())

    def connected_components(self):
        """
        Return the connected components.

        OUTPUT: list of connected constellations

        EXAMPLES::

            sage: c = Constellation(['(0,1)(2)', None, '(0,1)(2)'], connected=False)
            sage: cc = c.connected_components(); cc
            [Constellation of length 3 and degree 2
            g0 (0,1)
            g1 (0)(1)
            g2 (0,1),
            Constellation of length 3 and degree 1
            g0 (0)
            g1 (0)
            g2 (0)]
            sage: all(c2.is_connected() for c2 in cc)
            True

            sage: c = Constellation(['(0,1,2)', None], connected=False)
            sage: c.connected_components()
            [Constellation of length 2 and degree 3
            g0 (0,1,2)
            g1 (0,2,1)]
        """
        if self._connected:
            return [self]
        G = Graph()
        G.add_vertices(list(range(self.degree())))
        for p in self._g:
            G.add_edges(enumerate(p.domain()), loops=False)
        m = G.connected_components(sort=False)
        if len(m) == 1:
            return [self]
        for mm in m:
            mm.sort()
        m.sort()
        g = [[] for _ in repeat(None, len(m))]
        m_inv = [None] * self.degree()
        for t, mt in enumerate(m):
            for i, mti in enumerate(mt):
                m_inv[mti] = i
            for k in range(self.length()):
                tmp = [None] * len(mt)
                for i, mti in enumerate(mt):
                    tmp[i] = m_inv[self._g[k](mti)]
                g[t].append(tmp)
        return [Constellation(g=g[i], check=False) for i in range(len(m))]

    def _richcmp_(self, other, op):
        r"""
        Do the comparison.

        TESTS::

            sage: Constellation(['(0,1,2)', None]) == Constellation(['(0,1,2)', None])
            True
            sage: Constellation(['(0,1)','(0,2)',None]) == Constellation(['(0,1)',None,'(0,2)'])
            False

            sage: Constellation(['(0,1,2)', None]) != Constellation(['(0,1,2)', None])
            False
            sage: Constellation(['(0,1)','(0,2)',None]) != Constellation(['(0,1)',None,'(0,2)'])
            True

            sage: c1 = Constellation([[1,2,0],None])
            sage: c2 = Constellation([[2,0,1],None])
            sage: c1 < c2
            True
            sage: c2 > c1
            True
        """
        if not isinstance(other, Constellation_class):
            return op == op_NE
        if op == op_EQ:
            return self._g == other._g
        if op == op_NE:
            return self._g != other._g

        lx = self.length()
        rx = other.length()
        if lx != rx:
            return richcmp_not_equal(lx, rx, op)

        lx = self.degree()
        rx = other.degree()
        if lx != rx:
            return richcmp_not_equal(lx, rx, op)

        for i in range(self.length() - 1):
            lx = self._g[i]
            rx = other._g[i]
            if lx != rx:
                return richcmp_not_equal(lx, rx, op)
        return rich_to_bool(op, 0)

    def is_isomorphic(self, other, return_map=False):
        r"""
        Test of isomorphism.

        Return ``True`` if the constellations are isomorphic
        (*i.e.* related by a common conjugacy) and return the permutation that
        conjugate the two permutations if ``return_map`` is ``True`` in
        such a way that ``self.relabel(m) == other``.

        ALGORITHM:

        uses canonical labels obtained from the method :meth:`relabel`.

        EXAMPLES::

            sage: c = Constellation([[1,0,2],[2,1,0],[0,2,1],None])
            sage: d = Constellation([[2,1,0],[0,2,1],[1,0,2],None])
            sage: answer, mapping = c.is_isomorphic(d,return_map=True)
            sage: answer
            True
            sage: c.relabel(mapping) == d
            True
        """
        if return_map:
            if not (self.degree() == other.degree() and
                   self.length() == other.length()):
                return False, None
            sn, sn_map = self.relabel(return_map=True)
            on, on_map = other.relabel(return_map=True)
            if sn != on:
                return False, None
            return True, sn_map * ~on_map

        return (self.degree() == other.degree() and
                self.length() == other.length() and
                self.relabel() == other.relabel())

    def _repr_(self):
        r"""
        Return a string representation.

        EXAMPLES::

            sage: c = Constellation([[1,0,2],[2,1,0],[0,2,1],None])
            sage: c._repr_()
            'Constellation of length 4 and degree 3\ng0 (0,1)(2)\ng1 (0,2)(1)\ng2 (0)(1,2)\ng3 (0,2)(1)'
        """
        s = "Constellation of length {} and degree {}".format(self.length(),
                                                              self.degree())
        for i in range(self.length()):
            s += "\ng{} {}".format(i, self._g[i].cycle_string(True))
        return s

    def degree(self):
        r"""
        Return the degree of the constellation.

        The degree of a constellation is the number `n` that
        corresponds to the symmetric group `S(n)` in which the
        permutations of the constellation are defined.

        EXAMPLES::

            sage: c = Constellation([])
            sage: c.degree()
            0
            sage: c = Constellation(['(0,1)',None])
            sage: c.degree()
            2
            sage: c = Constellation(['(0,1)','(0,3,2)(1,5)',None,'(4,3,2,1)'])
            sage: c.degree()
            6

        TESTS::

            sage: parent(c.degree())
            Integer Ring
        """
        return self.parent()._degree

    def length(self):
        r"""
        Return the number of permutations.

        EXAMPLES::

            sage: c = Constellation(['(0,1)','(0,2)','(0,3)',None])
            sage: c.length()
            4
            sage: c = Constellation(['(0,1,3)',None,'(1,2)'])
            sage: c.length()
            3

        TESTS::

            sage: parent(c.length())
            Integer Ring
        """
        return Integer(len(self._g))

    def profile(self, i=None):
        r"""
        Return the profile of ``self``.

        The profile of a constellation is the tuple of partitions
        associated to the conjugacy classes of the permutations of the
        constellation.

        This is also called the passport.

        EXAMPLES::

            sage: c = Constellation(['(0,1,2)(3,4)','(0,3)',None])
            sage: c.profile()
            ([3, 2], [2, 1, 1, 1], [5])
        """
        if i is None:
            return tuple(self.profile(j) for j in range(self.length()))
        else:
            parts = [len(cy) for cy in self._g[i].cycle_tuples(True)]
            return Partition(sorted(parts, reverse=True))

    passport = profile

    # ACCESS TO INDIVIDUAL PERMUTATION

    def g(self, i=None):
        r"""
        Return the permutation `g_i` of the constellation.

        INPUT:

        - ``i`` -- integer or ``None`` (default)

        If ``None`` , return instead the list of all `g_i`.

        EXAMPLES::

            sage: c = Constellation(['(0,1,2)(3,4)','(0,3)',None])
            sage: c.g(0)
            (0,1,2)(3,4)
            sage: c.g(1)
            (0,3)
            sage: c.g(2)
            (0,4,3,2,1)
            sage: c.g()
            [(0,1,2)(3,4), (0,3), (0,4,3,2,1)]
        """
        from copy import copy
        if i is None:
            return copy(self._g)
        else:
            gi = self._g[i]
            return gi.parent()(gi)

    def relabel(self, perm=None, return_map=False):
        r"""
        Relabel ``self``.

        If ``perm`` is provided then relabel with respect to ``perm``. Otherwise
        use canonical labels. In that case, if ``return_map`` is provided, the
        return also the map used for canonical labels.

        Algorithm:

        the cycle for g(0) are adjacent and the cycle are joined with
        respect to the other permutations. The minimum is taken for
        all possible renumerotations.

        EXAMPLES::

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None]); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2)(3)
            g2 (0,1,3,2,4)
            sage: c2 = c.relabel(); c2
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,2)(3)(4)
            g2 (0,1,4,3,2)

        The map returned when the option ``return_map`` is set to
        ``True`` can be used to set the relabelling::

            sage: c3, perm = c.relabel(return_map=True)
            sage: c3 == c2 and c3 == c.relabel(perm=perm)
            True

            sage: S5 = SymmetricGroup(range(5))
            sage: d = c.relabel(S5([4,3,1,0,2])); d
            Constellation of length 3 and degree 5
            g0 (0,2,1)(3,4)
            g1 (0)(1)(2,3)(4)
            g2 (0,1,2,4,3)
            sage: d.is_isomorphic(c)
            True

        We check that after a random relabelling the new constellation is
        isomorphic to the initial one::

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None])
            sage: p = S5.random_element()
            sage: cc = c.relabel(perm=p)
            sage: cc.is_isomorphic(c)
            True

        Check that it works for "non standard" labels::

            sage: c = Constellation([(('a','b'),('c','d','e')),('b','d'), None])
            sage: c.relabel()
            Constellation of length 3 and degree 5
            g0 ('a','b')('c','d','e')
            g1 ('a')('b','c')('d')('e')
            g2 ('a','b','e','d','c')
        """
        if perm is not None:
            g = [[None] * self.degree() for _ in range(self.length())]
            for i in range(len(perm.domain())):
                for k in range(self.length()):
                    g[k][perm(i)] = perm(self._g[k](i))
            return Constellation(g=g, check=False, mutable=self.is_mutable())

        if return_map:
            try:
                return self._normal_form, self._normal_form_map
            except AttributeError:
                pass
        else:
            try:
                return self._normal_form
            except AttributeError:
                pass

        # compute canonical labels
        if not self.is_connected():
            raise ValueError("no canonical labels implemented for"
                             " non connected constellation")

        # get the permutations on {0, 1, ..., d-1}
        # compute the canonical labels
        # map it back to the domain
        # TODO: a lot of time is lost here!
        domain = list(self.parent()._sym.domain())
        index = {e: i for i, e in enumerate(domain)}
        g = [[index[gg(i)] for i in domain] for gg in self._g]
        c_win, m_win = perms_canonical_labels(g)
        c_win = [[domain[i] for i in gg] for gg in c_win]
        m_win = self.parent()._sym([domain[i] for i in m_win])
        c_win = self.parent()(c_win, mutable=False, check=False)

        if not self.is_mutable():
            self._normal_form = c_win
            self._normal_form_map = m_win

        c_win._normal_form = c_win
        c_win._normal_form_map = m_win

        if return_map:
            return c_win, m_win
        else:
            return c_win

    # BRAID GROUP ACTION

    def braid_group_action(self, i):
        r"""
        Act on ``self`` as the braid group generator that exchanges
        position `i` and `i+1`.

        INPUT:

        - ``i`` -- integer in `[0, n-1]` where `n` is the length of ``self``

        EXAMPLES::

            sage: sigma = lambda c, i: c.braid_group_action(i)

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None]); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2)(3)
            g2 (0,1,3,2,4)
            sage: sigma(c, 1)
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0,1,3,2,4)
            g2 (0,3)(1)(2)(4)

        Check the commutation relation::

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)','(2,5)(0,4)',None])
            sage: d = Constellation(['(0,1,3,5)','(2,3,4)','(0,3,5)',None])
            sage: c13 = sigma(sigma(c, 0), 2)
            sage: c31 = sigma(sigma(c, 2), 0)
            sage: c13 == c31
            True
            sage: d13 = sigma(sigma(d, 0), 2)
            sage: d31 = sigma(sigma(d, 2), 0)
            sage: d13 == d31
            True

        Check the braid relation::

            sage: c121 = sigma(sigma(sigma(c, 1), 2), 1)
            sage: c212 = sigma(sigma(sigma(c, 2), 1), 2)
            sage: c121 == c212
            True
            sage: d121 = sigma(sigma(sigma(d, 1), 2), 1)
            sage: d212 = sigma(sigma(sigma(d, 2), 1), 2)
            sage: d121 == d212
            True
        """
        if i < 0 or i >= self.length():
            txt = "i should be between 0 and {}"
            raise ValueError(txt.format(self.length() - 1))
        j = i + 1
        if j == self.length():   # wrap around the cylinder
            j = 0
        h = self.copy()
        si = self._g[i]
        sj = self._g[j]
        h._g[i] = sj
        h._g[j] = (~sj * si) * sj
        return h

    def braid_group_orbit(self):
        r"""
        Return the graph of the action of the braid group.

        The action is considered up to isomorphism of constellation.

        EXAMPLES::

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None]); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2)(3)
            g2 (0,1,3,2,4)
            sage: G = c.braid_group_orbit()
            sage: G.num_verts()
            4
            sage: G.num_edges()
            12
        """
        G = DiGraph(multiedges=True, loops=True)
        waiting = [self.relabel()]

        while waiting:
            c = waiting.pop()
            G.add_vertex(c)
            for i in range(self.length()):
                cc = self.braid_group_action(i).relabel()
                if cc not in G:
                    waiting.append(cc)
                G.add_edge(c, cc, i)
        return G


class Constellations_ld(UniqueRepresentation, Parent):
    r"""
    Constellations of given length and degree.

    EXAMPLES::

        sage: C = Constellations(2,3); C
        Connected constellations of length 2 and degree 3 on {1, 2, 3}
        sage: C([[2,3,1],[3,1,2]])
        Constellation of length 2 and degree 3
        g0 (1,2,3)
        g1 (1,3,2)
        sage: C.cardinality()
        2
        sage: Constellations(2,3,connected=False).cardinality()
        6
    """
    Element = Constellation_class

    def __init__(self, length, degree, sym=None, connected=True):
        """
        TESTS::

            sage: TestSuite(Constellations(length=6,degree=4)).run(skip='_test_cardinality')
        """
        from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
        Parent.__init__(self, category=FiniteEnumeratedSets())
        self._length = length
        self._degree = degree
        if self._length < 0:
            raise ValueError("length should be a nonnegative integer")
        if self._degree < 0:
            raise ValueError("degree should be a nonnegative integer")

        self._sym = sym

        self._connected = bool(connected)

    def is_empty(self):
        r"""
        Return whether this set of constellations is empty.

        EXAMPLES::

            sage: Constellations(2, 3).is_empty()
            False
            sage: Constellations(1, 2).is_empty()
            True
            sage: Constellations(1, 2, connected=False).is_empty()
            False
        """
        return self._connected and self._length == 1 and self._degree > 1

    def __contains__(self, elt):
        r"""
        TESTS::

            sage: C = Constellations(2, 3, connected=True)
            sage: D = Constellations(2, 3, connected=False)
            sage: e1 = [[3,1,2], None]
            sage: e2 = [[1,2,3], None]
            sage: C(e1) in C
            True
            sage: D(e1) in C
            True
            sage: D(e1) in D
            True
            sage: D(e2) in C
            False
            sage: D(e2) in D
            True

            sage: e1 in C and e1 in D
            True
            sage: e2 in C
            False
            sage: e2 in D
            True
        """
        if isinstance(elt, (tuple, list)):
            try:
                self(elt, check=True)
            except (ValueError, TypeError):
                return False
            else:
                return True
        elif not isinstance(elt, Constellation_class):
            return False
        return (elt.parent() is self or
                (elt.length() == self._length and
                 elt.degree() == self._degree and
                 (not self._connected or elt.is_connected())))

    def _repr_(self):
        """
        TESTS::

            sage: Constellations(3,3)._repr_()
            'Connected constellations of length 3 and degree 3 on {1, 2, 3}'
            sage: Constellations(3,3,connected=False)._repr_()
            'Constellations of length 3 and degree 3 on {1, 2, 3}'
        """
        s = "of length {} and degree {} on {}".format(self._length,
                                                      self._degree,
                                                      self._sym.domain())
        if self._connected:
            return "Connected constellations " + s
        else:
            return "Constellations " + s

    def __iter__(self):
        """
        Iterator over all constellations of given degree and length.

        EXAMPLES::

            sage: const = Constellations(3,3); const
            Connected constellations of length 3 and degree 3 on {1, 2, 3}
            sage: len([v for v in const])
            26

        One can check the first few terms of sequence :oeis:`220754`::

            sage: Constellations(4,1).cardinality()
            1
            sage: Constellations(4,2).cardinality()
            7
            sage: Constellations(4,3).cardinality()
            194
            sage: Constellations(4,4).cardinality()  # long time
            12858
        """
        from itertools import product

        if self._length == 1:
            if self._degree == 1:
                yield self([[0]])
            return

        S = self._sym
        for p in product(S, repeat=self._length - 1):
            if self._connected and not perms_are_connected(p, self._degree):
                continue
            yield self(list(p) + [None], check=False)

    def random_element(self, mutable=False):
        r"""
        Return a random element.

        This is found by trial and rejection, starting from
        a random list of permutations.

        EXAMPLES::

            sage: const = Constellations(3,3)
            sage: const.random_element()
            Constellation of length 3 and degree 3
            ...
            ...
            ...
            sage: c = const.random_element()
            sage: c.degree() == 3 and c.length() == 3
            True
        """
        from sage.groups.perm_gps.permgroup import PermutationGroup

        l = self._length
        d = self._degree
        Sd = self._sym

        g = [Sd.random_element() for _ in range(l - 1)]
        G = PermutationGroup(g)
        while not G.degree() == d or (self._connected and
                                      not G.is_transitive()):
            g = [Sd.random_element() for _ in range(l - 1)]
            G = PermutationGroup(g)

        return self([sigma.domain() for sigma in g] + [None], mutable=mutable)

    def _element_constructor_(self, *data, **options):
        r"""
        Build an element of ``self``.

        EXAMPLES::

            sage: C = Constellations(2,3)
            sage: C([[2,3,1],[3,1,2]])
            Constellation of length 2 and degree 3
            g0 (1,2,3)
            g1 (1,3,2)
            sage: C([[3,2,1],[3,2,1]])
            Traceback (most recent call last):
            ...
            ValueError: not connected
        """
        if len(data) == 1 and isinstance(data[0], (list, tuple)) and \
           len(data[0]) == self._length:
            g = list(data[0])
        else:
            g = list(data)

        if len(g) != self._length:
            raise ValueError("must be a list of length {}".format(self._length))

        if g.count(None) == 0:
            i = None
        elif g.count(None) == 1:
            i = g.index(None)
            del g[i]
        else:
            raise ValueError("at most one permutation can be None")

        g = [self._sym(w) for w in g]

        if i is not None:
            h = self._sym.one()
            for p in g[i:]:
                h *= p
            for p in g[:i]:
                h *= p
            g.insert(i, ~h)

        mutable = options.pop('mutable', False)
        if options.pop('check', True):
            c = self.element_class(self, g, self._connected, mutable, True)
            if c.degree() != self._degree:
                raise ValueError("degree is not {}".format(self._degree))
            if c.length() != self._length:
                raise ValueError("length is not {}".format(self._length))
            return c
        else:
            return self.element_class(self, g, self._connected, mutable, False)

    def _an_element_(self):
        r"""
        Return a constellation in ``self``.

        EXAMPLES::

            sage: Constellations(2, 3).an_element()
            Constellation of length 2 and degree 3
            g0 (1,3,2)
            g1 (1,2,3)

            sage: Constellations(3, 5,domain='abcde').an_element()
            Constellation of length 3 and degree 5
            g0 ('a','e','d','c','b')
            g1 ('a','b','c','d','e')
            g2 ('a')('b')('c')('d')('e')

            sage: Constellations(0, 0).an_element()
            Constellation of length 0 and degree 0

            sage: Constellations(1, 1).an_element()
            Constellation of length 1 and degree 1
            g0 (1)

            sage: Constellations(1, 2).an_element()
            Traceback (most recent call last):
            ...
            EmptySetError
        """
        if self.is_empty():
            from sage.categories.sets_cat import EmptySetError
            raise EmptySetError

        if self._degree == 0 and self._length == 0:
            return self([])
        elif self._length == 1:
            return self(self._sym.one())

        d = self._degree
        domain = self._sym.domain().list()
        if self._connected:
            g = [[domain[d - 1]] + domain[:d - 1], domain[1:] + [domain[0]]]
            g += [domain[:]] * (self._length - 2)
        else:
            g = [domain[:]] * self._length
        return self(g)

    def braid_group_action(self):
        r"""
        Return a list of graphs that corresponds to the braid group action on
        ``self`` up to isomorphism.

        OUTPUT: list of graphs

        EXAMPLES::

            sage: C = Constellations(3,3)
            sage: C.braid_group_action()
            [Looped multi-digraph on 3 vertices,
             Looped multi-digraph on 1 vertex,
             Looped multi-digraph on 3 vertices]
        """
        G = []
        for c in self:
            c = c.relabel()
            if any(c in g for g in G):
                continue
            G.append(c.braid_group_orbit())
        return G

    def braid_group_orbits(self):
        r"""
        Return the orbits under the action of braid group.

        EXAMPLES::

            sage: C = Constellations(3,3)
            sage: O = C.braid_group_orbits()
            sage: len(O)
            3
            sage: [x.profile() for x in O[0]]
            [([1, 1, 1], [3], [3]), ([3], [1, 1, 1], [3]), ([3], [3], [1, 1, 1])]
            sage: [x.profile() for x in O[1]]
            [([3], [3], [3])]
            sage: [x.profile() for x in O[2]]
            [([2, 1], [2, 1], [3]), ([2, 1], [3], [2, 1]), ([3], [2, 1], [2, 1])]
        """
        return [g.vertices(sort=True) for g in self.braid_group_action()]


class Constellations_p(UniqueRepresentation, Parent):
    r"""
    Constellations with fixed profile.

    EXAMPLES::

        sage: C = Constellations([[3,1],[3,1],[2,2]]); C
        Connected constellations with profile ([3, 1], [3, 1], [2, 2]) on {1, 2, 3, 4}
        sage: C.cardinality()
        24
        sage: C.first()
        Constellation of length 3 and degree 4
        g0 (1)(2,3,4)
        g1 (1,2,3)(4)
        g2 (1,2)(3,4)
        sage: C.last()
        Constellation of length 3 and degree 4
        g0 (1,4,3)(2)
        g1 (1,4,2)(3)
        g2 (1,2)(3,4)

    Note that the cardinality can also be computed using characters of the
    symmetric group (Frobenius formula)::

        sage: P = Partitions(4)
        sage: p1 = Partition([3,1])
        sage: p2 = Partition([3,1])
        sage: p3 = Partition([2,2])
        sage: i1 = P.cardinality() - P.rank(p1) - 1
        sage: i2 = P.cardinality() - P.rank(p2) - 1
        sage: i3 = P.cardinality() - P.rank(p3) - 1
        sage: s = 0
        sage: for c in SymmetricGroup(4).irreducible_characters():
        ....:     v = c.values()
        ....:     s += v[i1] * v[i2] * v[i3] / v[0]
        sage: c1 = p1.conjugacy_class_size()
        sage: c2 = p2.conjugacy_class_size()
        sage: c3 = p3.conjugacy_class_size()
        sage: c1 * c2 * c3 / factorial(4)**2 * s
        1

    The number obtained above is up to isomorphism. And we can check::

        sage: len(C.isomorphism_representatives())
        1
    """

    def __init__(self, profile, domain=None, connected=True):
        r"""
        OPTIONS:

        - ``profile`` -- list of integer partitions of the same integer

        - ``connected`` -- boolean (default: ``True``); whether we consider
          only connected constellations

        TESTS::

            sage: C = Constellations([(3,1),(3,1),(2,2)])
            sage: TestSuite(C).run()
        """
        l = Integer(len(profile))
        d = Integer(sum(profile[0]))
        for p in profile:
            if sum(p) != d:
                raise ValueError("all partition in the passport should "
                                 "have the same sum.")
        if domain is None:
            sym = SymmetricGroup(d)
        else:
            sym = SymmetricGroup(domain)
            if len(sym.domain()) != d:
                raise ValueError("the size of the domain should be equal to the degree")

        self._cd = Constellations_ld(l, d, sym, connected)
        self._profile = profile
        from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def _repr_(self):
        r"""
        TESTS::

            sage: Constellations(profile=[[3,2,1],[3,3],[3,3]])
            Connected constellations with profile ([3, 2, 1], [3, 3], [3, 3]) on {1, 2, 3, 4, 5, 6}
        """
        s = "with profile {} on {}".format(self._profile,
                                           self._cd._sym.domain())
        if self._cd._connected:
            return "Connected constellations " + s
        return "Constellations " + s

    def isomorphism_representatives(self):
        r"""
        Return a set of isomorphism representative of ``self``.

        EXAMPLES::

            sage: C = Constellations([[5], [4,1], [3,2]])
            sage: C.cardinality()
            240
            sage: ir = sorted(C.isomorphism_representatives())
            sage: len(ir)
            2
            sage: ir[0]
            Constellation of length 3 and degree 5
            g0 (1,2,3,4,5)
            g1 (1)(2,3,4,5)
            g2 (1,5,3)(2,4)
            sage: ir[1]
            Constellation of length 3 and degree 5
            g0 (1,2,3,4,5)
            g1 (1)(2,5,3,4)
            g2 (1,5)(2,3,4)
        """
        result = set()
        for c in self:
            cc = c.relabel()
            if cc not in result:
                result.add(cc)
        return result

    def _element_constructor_(self, *data, **options):
        r"""
        Build an element of ``self``.

        TESTS::

            sage: C = Constellations([(3,1),(3,1),(2,2)])
            sage: c = C([(2,3,4),(1,2,3),((1,2),(3,4))]); c
            Constellation of length 3 and degree 4
            g0 (1)(2,3,4)
            g1 (1,2,3)(4)
            g2 (1,2)(3,4)
            sage: C([(1,2,3),(3,2,4),None])
            Traceback (most recent call last):
            ...
            ValueError: wrong profile
        """
        c = self._cd(*data, **options)
        if options.get('check', True) and c.profile() != self._profile:
            raise ValueError("wrong profile")
        return c

    def __iter__(self):
        r"""
        Iterator of the elements in ``self``.

        TESTS::

            sage: C = Constellations([(3,1),(3,1),(2,2)])
            sage: for c in C: print(c)
            Constellation of length 3 and degree 4
            g0 (1)(2,3,4)
            g1 (1,2,3)(4)
            g2 (1,2)(3,4)
            Constellation of length 3 and degree 4
            g0 (1)(2,3,4)
            g1 (1,4,2)(3)
            g2 (1,4)(2,3)
            ...
            Constellation of length 3 and degree 4
            g0 (1,4,3)(2)
            g1 (1,2,3)(4)
            g2 (1,4)(2,3)
            Constellation of length 3 and degree 4
            g0 (1,4,3)(2)
            g1 (1,4,2)(3)
            g2 (1,2)(3,4)

            sage: C = Constellations([(3,1),(3,1),(2,2)], domain='abcd')
            sage: for c in sorted(C): print(c)
            Constellation of length 3 and degree 4
            g0 ('a')('b','c','d')
            g1 ('a','b','c')('d')
            g2 ('a','b')('c','d')
            ...
            Constellation of length 3 and degree 4
            g0 ('a','d','c')('b')
            g1 ('a','d','b')('c')
            g2 ('a','b')('c','d')
        """
        if self._cd._length == 1:
            if self._cd._degree == 1:
                yield self([[0]])
            return

        S = self._cd._sym
        profile = list(self._profile)[:-1]
        for p in product(*[S.conjugacy_class(pi) for pi in profile]):
            if self._cd._connected and not perms_are_connected(p, self._cd._degree):
                continue
            c = self._cd(list(p) + [None], check=False)
            if c.profile() == self._profile:
                yield c

# *************************************************************************
#                          auxiliary functions
# *************************************************************************


def perm_sym_domain(g):
    r"""
    Return the domain of a single permutation (before initialization).

    EXAMPLES::

        sage: from sage.combinat.constellation import perm_sym_domain
        sage: perm_sym_domain([1,2,3,4])
        {1, 2, 3, 4}
        sage: perm_sym_domain(((1,2),(0,4)))
        {0, 1, 2, 4}
        sage: sorted(perm_sym_domain('(1,2,0,5)'))
        [0, 1, 2, 5]
    """
    if isinstance(g, (tuple, list)):
        if isinstance(g[0], tuple):
            return set().union(*g)
        else:
            return set(g)
    elif isinstance(g, str):  # perms given as strings of cycles
        assert g.startswith('(') and g.endswith(')')
        domain = set().union(*[a for cyc in g[1:-1].split(')(')
                               for a in cyc.split(',')])
        if all(s.isdigit() for s in domain):
            return [int(x) for x in domain]
        else:
            return domain
    elif parent(g) in Groups:
        return g.domain()
    else:
        raise TypeError


def perms_sym_init(g, sym=None):
    r"""
    Initialize a list of permutations (in the same symmetric group).

    OUTPUT:

    - ``sym`` -- a symmetric group

    - ``gg`` -- list of permutations

    EXAMPLES::

        sage: from sage.combinat.constellation import perms_sym_init
        sage: S, g = perms_sym_init([[0,2,1,3], [1,3,2,0]])
        sage: S.domain()
        {0, 1, 2, 3}
        sage: g
        [(1,2), (0,1,3)]

        sage: S, g = perms_sym_init(['(2,1)', '(0,3)'])
        sage: S.domain()
        {0, 1, 2, 3}
        sage: g
        [(1,2), (0,3)]

        sage: S, g = perms_sym_init([(1,0), (2,1)])
        sage: S.domain()
        {0, 1, 2}
        sage: g
        [(0,1), (1,2)]

        sage: S, g = perms_sym_init([((1,0),(2,3)), '(0,1,4)'])
        sage: S.domain()
        {0, 1, 2, 3, 4}
        sage: g
        [(0,1)(2,3), (0,1,4)]
    """
    if g is None or len(g) == 0:
        if sym is None:
            sym = SymmetricGroup(0)
        return sym, [sym([])]

    if sym is None:
        domain = set().union(*[perm_sym_domain(gg) for gg in g])
        if all(isinstance(s, (int, Integer)) and s > 0
               for s in domain):
            domain = max(domain)
        else:
            domain = sorted(domain)
        sym = SymmetricGroup(domain)

    try:
        return sym, [sym(u) for u in g]
    except (ValueError, TypeError):
        return sym, None


def perms_are_connected(g, n):
    """
    Check that the action of the generated group is transitive.

    INPUT:

    - ``g`` -- list of permutations of `[0, n-1]` (in a SymmetricGroup)

    - ``n`` -- integer

    EXAMPLES::

        sage: from sage.combinat.constellation import perms_are_connected
        sage: S = SymmetricGroup(range(3))
        sage: perms_are_connected([S([0,1,2]),S([0,2,1])],3)
        False
        sage: perms_are_connected([S([0,1,2]),S([1,2,0])],3)
        True
    """
    G = Graph()
    if g:
        G.add_vertices(g[0].domain())
    for p in g:
        G.add_edges(p.dict().items(), loops=False)
    return G.is_connected()


def perms_canonical_labels_from(x, y, j0, verbose=False):
    r"""
    Return canonical labels for ``x``, ``y`` that starts at ``j0``.

    .. WARNING::

        The group generated by ``x`` and the elements of ``y`` should be
        transitive.

    INPUT:

    - ``x`` -- list; a permutation of `[0, ..., n]` as a list

    - ``y`` -- list of permutations of `[0, ..., n]` as a list of lists

    - ``j0`` -- an index in [0, ..., n]

    OUTPUT: mapping: a permutation that specify the new labels

    EXAMPLES::

        sage: from sage.combinat.constellation import perms_canonical_labels_from
        sage: perms_canonical_labels_from([0,1,2],[[1,2,0]], 0)
        [0, 1, 2]
        sage: perms_canonical_labels_from([1,0,2], [[2,0,1]], 0)
        [0, 1, 2]
        sage: perms_canonical_labels_from([1,0,2], [[2,0,1]], 1)
        [1, 0, 2]
        sage: perms_canonical_labels_from([1,0,2], [[2,0,1]], 2)
        [2, 1, 0]
    """
    n = len(x)

    k = 0
    mapping = [None] * n
    waiting = [[] for _ in repeat(None, len(y))]

    while k < n:
        if verbose:
            print("complete from {}".format(j0))
        # initialize at j0
        mapping[j0] = k
        waiting[0].append(j0)
        k += 1
        # complete x cycle from j0
        j = x[j0]
        while j != j0:
            mapping[j] = k
            waiting[0].append(j)
            k += 1
            j = x[j]
        if verbose:
            print("completed cycle mapping = {}".format(mapping))

        # find another guy
        if verbose:
            print("try to find somebody in {}".format(waiting))
        l = 0
        while l < len(waiting):
            i = 0
            while i < len(waiting[l]):
                j1 = waiting[l][i]
                if mapping[y[l][j1]] is None:
                    break
                i += 1

            if i == len(waiting[l]):  # not found: go further in waiting
                if l < len(waiting) - 1:
                    waiting[l + 1].extend(waiting[l])
                waiting[l] = []
                l += 1
                i = 0

            else:  # found: complete cycle from new guy
                j0 = y[l][j1]
                if l < len(waiting) - 1:
                    waiting[l + 1].extend(waiting[l][:i + 1])
                del waiting[l][:i + 1]
                break

    return mapping


def perm_invert(p):
    """
    Return the inverse of the permutation `p`.

    INPUT:

    - ``p`` -- a permutation of {0,..,n-1} given by a list of values

    OUTPUT: a permutation of {0,..,n-1} given by a list of values

    EXAMPLES::

        sage: from sage.combinat.constellation import perm_invert
        sage: perm_invert([3,2,0,1])
        [2, 3, 1, 0]
    """
    q = [None] * len(p)
    for i, j in enumerate(p):
        q[j] = i
    return q


def perm_conjugate(p, s):
    """
    Return the conjugate of the permutation `p` by the permutation `s`.

    INPUT:

    - ``p``, ``s`` -- two permutations of {0,..,n-1} given by lists of values

    OUTPUT: a permutation of {0,..,n-1} given by a list of values

    EXAMPLES::

        sage: from sage.combinat.constellation import perm_conjugate
        sage: perm_conjugate([3,1,2,0], [3,2,0,1])
        [0, 3, 2, 1]
    """
    q = [None] * len(p)
    for i in range(len(p)):
        q[s[i]] = s[p[i]]
    return q


def perms_canonical_labels(p, e=None):
    """
    Relabel a list with a common conjugation such that two conjugated
    lists are relabeled the same way.

    INPUT:

    - ``p`` -- list of at least 2 permutations

    - ``e`` -- ``None`` or a list of integer in the domain of the
      permutations. If provided, then the renumbering algorithm is
      only performed from the elements of ``e``.

    OUTPUT: a pair made of a list of permutations (as a list of lists) and a
    list that corresponds to the conjugacy used.

    EXAMPLES::

        sage: from sage.combinat.constellation import perms_canonical_labels
        sage: l0 = [[2,0,3,1], [3,1,2,0], [0,2,1,3]]
        sage: l, m = perms_canonical_labels(l0); l
        [[1, 2, 3, 0], [0, 3, 2, 1], [2, 1, 0, 3]]

        sage: S = SymmetricGroup(range(4))
        sage: [~S(m) * S(u) * S(m) for u in l0] == list(map(S, l))
        True

        sage: perms_canonical_labels([])
        Traceback (most recent call last):
        ...
        ValueError: input must have length >= 2
    """
    if not len(p) > 1:
        raise ValueError('input must have length >= 2')

    n = len(p[0])

    c_win = None
    m_win = list(range(n))

    x = p[0]
    y = p[1:]

    if e is None:
        e = list(range(n))

    # get canonical label from i in to_test and compare
    while e:
        i = e.pop()
        m_test = perms_canonical_labels_from(x, y, i)
        c_test = [perm_conjugate(u, m_test) for u in p]
        if c_win is None or c_test < c_win:
            c_win = c_test
            m_win = m_test

    return c_win, m_win



# sage_setup: distribution = sagemath-combinat
r"""
Partitions, tableaux, and others

Partitions
----------

- :ref:`sage.combinat.partition`
- :ref:`sage.combinat.skew_partition`
- :ref:`sage.combinat.partition_tuple`
- :ref:`sage.combinat.superpartition`
- :ref:`sage.combinat.tableau`
- :ref:`sage.combinat.tableau_tuple`
- :ref:`sage.combinat.skew_tableau`
- :ref:`sage.combinat.ribbon`
- :ref:`sage.combinat.ribbon_tableau`
- :ref:`sage.combinat.k_tableau`
- :ref:`sage.combinat.shifted_primed_tableau`

- :ref:`sage.combinat.tableau_residues`

RSK
---

- :ref:`sage.combinat.rsk`
- :ref:`sage.combinat.growth`
"""



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Free pre-Lie algebras

AUTHORS:

- Florent Hivert, Frédéric Chapoton (2011)
"""

# ****************************************************************************
#       Copyright (C) 2010-2015 Florent Hivert <Florent.Hivert@lri.fr>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from itertools import product

from sage.categories.magmatic_algebras import MagmaticAlgebras
from sage.categories.lie_algebras import LieAlgebras
from sage.categories.magmas import Magmas
from sage.categories.pushout import (ConstructionFunctor,
                                     CompositeConstructionFunctor,
                                     IdentityConstructionFunctor)
from sage.categories.rings import Rings
from sage.categories.functor import Functor

from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.integer_vector import IntegerVectors
from sage.combinat.words.alphabet import Alphabet
from sage.combinat.rooted_tree import (RootedTrees, RootedTree,
                                       LabelledRootedTrees,
                                       LabelledRootedTree)
from sage.combinat.grossman_larson_algebras import GrossmanLarsonAlgebra, ROOT

from sage.misc.lazy_attribute import lazy_attribute
from sage.misc.cachefunc import cached_method
from sage.functions.other import factorial

from sage.sets.family import Family
from sage.structure.coerce_exceptions import CoercionException
from sage.rings.infinity import Infinity


class FreePreLieAlgebra(CombinatorialFreeModule):
    r"""
    The free pre-Lie algebra.

    Pre-Lie algebras are non-associative algebras, where the product `*`
    satisfies

    .. MATH::

        (x * y) * z - x * (y * z) = (x * z) * y - x * (z * y).

    We use here the convention where the associator

    .. MATH::

        (x, y, z) := (x * y) * z - x * (y * z)

    is symmetric in its two rightmost arguments. This is sometimes called
    a right pre-Lie algebra.

    They have appeared in numerical analysis and deformation theory.

    The free Pre-Lie algebra on a given set `E` has an explicit
    description using rooted trees, just as the free associative algebra
    can be described using words. The underlying vector space has a basis
    indexed by finite rooted trees endowed with a map from their vertices
    to `E`. In this basis, the product of two (decorated) rooted trees `S
    * T` is the sum over vertices of `S` of the rooted tree obtained by
    adding one edge from the root of `T` to the given vertex of `S`. The
    root of these trees is taken to be the root of `S`. The free pre-Lie
    algebra can also be considered as the free algebra over the PreLie operad.

    .. WARNING::

        The usual binary operator ``*`` can be used for the pre-Lie product.
        Beware that it but must be parenthesized properly, as the pre-Lie
        product is not associative. By default, a multiple product will be
        taken with left parentheses.

    EXAMPLES::

        sage: F = algebras.FreePreLie(ZZ, 'xyz')
        sage: x,y,z = F.gens()
        sage: (x * y) * z
        B[x[y[z[]]]] + B[x[y[], z[]]]
        sage: (x * y) * z - x * (y * z) == (x * z) * y - x * (z * y)
        True

    The free pre-Lie algebra is non-associative::

        sage: x * (y * z) == (x * y) * z
        False

    The default product is with left parentheses::

        sage: x * y * z == (x * y) * z
        True
        sage: x * y * z * x == ((x * y) * z) * x
        True

    The NAP product as defined in [Liv2006]_ is also implemented on the same
    vector space::

        sage: N = F.nap_product
        sage: N(x*y,z*z)
        B[x[y[], z[z[]]]]

    When ``None`` is given as input, unlabelled trees are used instead::

        sage: F1 = algebras.FreePreLie(QQ, None)
        sage: w = F1.gen(0); w
        B[[]]
        sage: w * w * w * w
        B[[[[[]]]]] + B[[[[], []]]] + 3*B[[[], [[]]]] + B[[[], [], []]]

    However, it is equally possible to use labelled trees instead::

        sage: F1 = algebras.FreePreLie(QQ, 'q')
        sage: w = F1.gen(0); w
        B[q[]]
        sage: w * w * w * w
        B[q[q[q[q[]]]]] + B[q[q[q[], q[]]]] + 3*B[q[q[], q[q[]]]] + B[q[q[], q[], q[]]]

    The set `E` can be infinite::

        sage: F = algebras.FreePreLie(QQ, ZZ)
        sage: w = F.gen(1); w
        B[1[]]
        sage: x = F.gen(2); x
        B[-1[]]
        sage: y = F.gen(3); y
        B[2[]]
        sage: w*x
        B[1[-1[]]]
        sage: (w*x)*y
        B[1[-1[2[]]]] + B[1[-1[], 2[]]]
        sage: w*(x*y)
        B[1[-1[2[]]]]

    Elements of a free pre-Lie algebra can be lifted to the universal
    enveloping algebra of the associated Lie algebra. The universal
    enveloping algebra is the Grossman-Larson Hopf algebra::

        sage: F = algebras.FreePreLie(QQ,'abc')
        sage: a,b,c = F.gens()
        sage: (a*b+b*c).lift()
        B[#[a[b[]]]] + B[#[b[c[]]]]

    .. NOTE::

        Variables names can be ``None``, a list of strings, a string
        or an integer. When ``None`` is given, unlabelled rooted
        trees are used. When a single string is given, each letter is taken
        as a variable. See
        :func:`sage.combinat.words.alphabet.build_alphabet`.

    .. WARNING::

        Beware that the underlying combinatorial free module is based
        either on ``RootedTrees`` or on ``LabelledRootedTrees``, with no
        restriction on the labellings. This means that all code calling
        the :meth:`basis` method would not give meaningful results, since
        :meth:`basis` returns many "chaff" elements that do not belong to
        the algebra.

    REFERENCES:

    - [ChLi]_

    - [Liv2006]_
    """
    @staticmethod
    def __classcall_private__(cls, R, names=None):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: F1 = algebras.FreePreLie(QQ, 'xyz')
            sage: F2 = algebras.FreePreLie(QQ, 'x,y,z')
            sage: F3 = algebras.FreePreLie(QQ, ['x','y','z'])
            sage: F4 = algebras.FreePreLie(QQ, Alphabet('xyz'))
            sage: F1 is F2 and F1 is F3 and F1 is F4
            True
        """
        if names is not None:
            if isinstance(names, str) and ',' in names:
                names = [u for u in names if u != ',']
            names = Alphabet(names)

        if R not in Rings():
            raise TypeError("argument R must be a ring")

        return super().__classcall__(cls, R, names)

    def __init__(self, R, names=None):
        """
        Initialize ``self``.

        TESTS::

            sage: A = algebras.FreePreLie(QQ, '@'); A
            Free PreLie algebra on one generator ['@'] over Rational Field
            sage: TestSuite(A).run()

            sage: A = algebras.FreePreLie(QQ, None); A
            Free PreLie algebra on one generator ['o'] over Rational Field

            sage: F = algebras.FreePreLie(QQ, 'xy')
            sage: TestSuite(F).run() # long time
        """
        if names is None:
            Trees = RootedTrees()
            key = RootedTree.sort_key
            self._alphabet = Alphabet(['o'])
        else:
            Trees = LabelledRootedTrees()
            key = LabelledRootedTree.sort_key
            self._alphabet = names
        # Here one would need LabelledRootedTrees(names)
        # so that one can restrict the labels to some fixed set

        cat = MagmaticAlgebras(R).WithBasis().Graded() & LieAlgebras(R).WithBasis().Graded()
        CombinatorialFreeModule.__init__(self, R, Trees,
                                         latex_prefix='',
                                         sorting_key=key,
                                         category=cat)

    def variable_names(self):
        r"""
        Return the names of the variables.

        EXAMPLES::

            sage: R = algebras.FreePreLie(QQ, 'xy')
            sage: R.variable_names()
            {'x', 'y'}

            sage: R = algebras.FreePreLie(QQ, None)
            sage: R.variable_names()
            {'o'}
        """
        return self._alphabet

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: algebras.FreePreLie(QQ, '@')  # indirect doctest
            Free PreLie algebra on one generator ['@'] over Rational Field

            sage: algebras.FreePreLie(QQ, ZZ)  # indirect doctest
            Free PreLie algebra on generators indexed by Integer Ring
            over Rational Field

            sage: enum = EnumeratedSets().Infinite().example()
            sage: algebras.FreePreLie(QQ, enum)  # indirect doctest
            Free PreLie algebra on generators indexed by An example of an
            infinite enumerated set: the nonnegative integers
            over Rational Field
        """
        n = self.algebra_generators().cardinality()
        finite = bool(n < Infinity)
        if not finite:
            gen = "generators indexed by"
        elif n == 1:
            gen = "one generator"
        else:
            gen = "{} generators".format(n)
        s = "Free PreLie algebra on {} {} over {}"
        if finite:
            try:
                return s.format(gen, self._alphabet.list(), self.base_ring())
            except NotImplementedError:
                return s.format(gen, self._alphabet, self.base_ring())
        else:
            return s.format(gen, self._alphabet, self.base_ring())

    def gen(self, i):
        r"""
        Return the ``i``-th generator of the algebra.

        INPUT:

        - ``i`` -- integer

        EXAMPLES::

            sage: F = algebras.FreePreLie(ZZ, 'xyz')
            sage: F.gen(0)
            B[x[]]

            sage: F.gen(4)
            Traceback (most recent call last):
            ...
            IndexError: argument i (= 4) must be between 0 and 2
        """
        G = self.algebra_generators()
        n = G.cardinality()
        if i < 0 or not i < n:
            m = "argument i (= {}) must be between 0 and {}".format(i, n - 1)
            raise IndexError(m)
        return G[G.keys().unrank(i)]

    @cached_method
    def algebra_generators(self):
        r"""
        Return the generators of this algebra.

        These are the rooted trees with just one vertex.

        EXAMPLES::

            sage: A = algebras.FreePreLie(ZZ, 'fgh'); A
            Free PreLie algebra on 3 generators ['f', 'g', 'h']
             over Integer Ring
            sage: list(A.algebra_generators())
            [B[f[]], B[g[]], B[h[]]]

            sage: A = algebras.FreePreLie(QQ, ['x1','x2'])
            sage: list(A.algebra_generators())
            [B[x1[]], B[x2[]]]
        """
        Trees = self.basis().keys()
        return Family(self._alphabet, lambda a: self.monomial(Trees([], a)))

    def change_ring(self, R):
        """
        Return the free pre-Lie algebra in the same variables over `R`.

        INPUT:

        - ``R`` -- a ring

        EXAMPLES::

            sage: A = algebras.FreePreLie(ZZ, 'fgh')
            sage: A.change_ring(QQ)
            Free PreLie algebra on 3 generators ['f', 'g', 'h'] over
            Rational Field
        """
        return FreePreLieAlgebra(R, names=self.variable_names())

    def gens(self) -> tuple:
        """
        Return the generators of ``self`` (as an algebra).

        EXAMPLES::

            sage: A = algebras.FreePreLie(ZZ, 'fgh')
            sage: A.gens()
            (B[f[]], B[g[]], B[h[]])
        """
        return tuple(self.algebra_generators())

    def degree_on_basis(self, t):
        """
        Return the degree of a rooted tree in the free Pre-Lie algebra.

        This is the number of vertices.

        EXAMPLES::

            sage: A = algebras.FreePreLie(QQ, None)
            sage: RT = A.basis().keys()
            sage: A.degree_on_basis(RT([RT([])]))
            2
        """
        return t.node_number()

    def _an_element_(self):
        """
        Return an element of ``self``.

        EXAMPLES::

            sage: A = algebras.FreePreLie(QQ, 'xy')
            sage: A.an_element()
            B[x[x[x[x[]]]]] + B[x[x[], x[x[]]]]
        """
        o = self.gen(0)
        return (o * o) * (o * o)

    def some_elements(self):
        """
        Return some elements of the free pre-Lie algebra.

        EXAMPLES::

            sage: A = algebras.FreePreLie(QQ, None)
            sage: A.some_elements()
            [B[[]], B[[[]]], B[[[[[]]]]] + B[[[], [[]]]], B[[[[]]]] + B[[[], []]], B[[[]]]]

        With several generators::

            sage: A = algebras.FreePreLie(QQ, 'xy')
            sage: A.some_elements()
            [B[x[]],
             B[x[x[]]],
             B[x[x[x[x[]]]]] + B[x[x[], x[x[]]]],
             B[x[x[x[]]]] + B[x[x[], x[]]],
             B[x[x[y[]]]] + B[x[x[], y[]]]]
        """
        o = self.gen(0)
        x = o * o
        y = o
        G = self.algebra_generators()
        # Take only the first 3 generators, otherwise the final element is too big
        if G.cardinality() < 3:
            for w in G:
                y = y * w
        else:
            K = G.keys()
            for i in range(3):
                y = y * G[K.unrank(i)]
        return [o, x, x * x, x * o, y]

    def product_on_basis(self, x, y):
        """
        Return the pre-Lie product of two trees.

        This is the sum over all graftings of the root of `y` over a vertex
        of `x`. The root of the resulting trees is the root of `x`.

        .. SEEALSO::

            :meth:`pre_Lie_product`

        EXAMPLES::

            sage: A = algebras.FreePreLie(QQ, None)
            sage: RT = A.basis().keys()
            sage: x = RT([RT([])])
            sage: A.product_on_basis(x, x)
            B[[[[[]]]]] + B[[[], [[]]]]
        """
        return self.sum(self.basis()[u] for u in x.graft_list(y))

    pre_Lie_product_on_basis = product_on_basis

    @lazy_attribute
    def pre_Lie_product(self):
        """
        Return the pre-Lie product.

        .. SEEALSO::

            :meth:`pre_Lie_product_on_basis`

        EXAMPLES::

            sage: A = algebras.FreePreLie(QQ, None)
            sage: RT = A.basis().keys()
            sage: x = A(RT([RT([])]))
            sage: A.pre_Lie_product(x, x)
            B[[[[[]]]]] + B[[[], [[]]]]
        """
        plb = self.pre_Lie_product_on_basis
        return self._module_morphism(self._module_morphism(plb, position=0,
                                                           codomain=self),
                                     position=1)

    def bracket_on_basis(self, x, y):
        r"""
        Return the Lie bracket of two trees.

        This is the commutator `[x, y] = x * y - y * x` of the pre-Lie product.

        .. SEEALSO::

            :meth:`pre_Lie_product_on_basis`

        EXAMPLES::

            sage: A = algebras.FreePreLie(QQ, None)
            sage: RT = A.basis().keys()
            sage: x = RT([RT([])])
            sage: y = RT([x])
            sage: A.bracket_on_basis(x, y)
            -B[[[[], [[]]]]] + B[[[], [[[]]]]] - B[[[[]], [[]]]]
        """
        return self.product_on_basis(x, y) - self.product_on_basis(y, x)

    def nap_product_on_basis(self, x, y):
        """
        Return the NAP product of two trees.

        This is the grafting of the root of `y` over the root
        of `x`. The root of the resulting tree is the root of `x`.

        .. SEEALSO::

            :meth:`nap_product`

        EXAMPLES::

            sage: A = algebras.FreePreLie(QQ, None)
            sage: RT = A.basis().keys()
            sage: x = RT([RT([])])
            sage: A.nap_product_on_basis(x, x)
            B[[[], [[]]]]
        """
        return self.basis()[x.graft_on_root(y)]

    @lazy_attribute
    def nap_product(self):
        """
        Return the NAP product.

        .. SEEALSO::

            :meth:`nap_product_on_basis`

        EXAMPLES::

            sage: A = algebras.FreePreLie(QQ, None)
            sage: RT = A.basis().keys()
            sage: x = A(RT([RT([])]))
            sage: A.nap_product(x, x)
            B[[[], [[]]]]
        """
        npb = self.nap_product_on_basis
        return self._module_morphism(self._module_morphism(npb,
                                                           position=0,
                                                           codomain=self),
                                     position=1)

    def corolla(self, x, y, n, N):
        """
        Return the corolla obtained with ``x`` as root and ``y`` as leaves.

        INPUT:

        - ``x``, ``y`` -- two elements
        - ``n`` -- integer; width of the corolla
        - ``N`` -- integer; truncation order (up to order ``N`` included)

        OUTPUT:

        the sum over all possible ways to graft ``n`` copies of ``y``
        on top of ``x`` (with at most ``N`` vertices in total)

        This operation can be defined by induction starting from the
        pre-Lie product.

        EXAMPLES::

            sage: A = algebras.FreePreLie(QQ)
            sage: a = A.gen(0)
            sage: b = A.corolla(a,a,1,4); b
            B[[[]]]
            sage: A.corolla(b,b,2,7)
            B[[[[[]], [[]]]]] + 2*B[[[[]], [[[]]]]] + B[[[], [[]], [[]]]]

            sage: A = algebras.FreePreLie(QQ, 'o')
            sage: a = A.gen(0)
            sage: b = A.corolla(a,a,1,4)

            sage: A = algebras.FreePreLie(QQ,'ab')
            sage: a, b = A.gens()
            sage: A.corolla(a,b,1,4)
            B[a[b[]]]
            sage: A.corolla(b,a,3,4)
            B[b[a[], a[], a[]]]

            sage: A.corolla(a+b,a+b,2,4)
            B[a[a[], a[]]] + 2*B[a[a[], b[]]] + B[a[b[], b[]]] + B[b[a[], a[]]] +
            2*B[b[a[], b[]]] + B[b[b[], b[]]]

        TESTS::

            sage: A = algebras.FreePreLie(QQ,'ab')
            sage: a, b = A.gens()
            sage: A.corolla(a,A.zero(),2,2)
            0
        """
        if not x or not y:
            return self.zero()

        basering = self.base_ring()
        vx = x.valuation()
        vy = y.valuation()
        min_deg = vy * n + vx
        if min_deg > N:
            return self.zero()

        try:
            self.gen(0).support()[0].label()
            labels = True
        except AttributeError:
            labels = False

        deg_x = x.maximal_degree()
        deg_y = y.maximal_degree()
        max_x = min(deg_x, N - n * vy)
        max_y = min(deg_y, N - vx - (n - 1) * vy)
        xx = x.truncate(max_x + 1)
        yy = y.truncate(max_y + 1)

        y_homog = {i: list(yy.homogeneous_component(i))
                   for i in range(vy, max_y + 1)}
        resu = self.zero()
        for k in range(min_deg, N + 1):   # total degree of (x ; y, y, y, y)
            for mx, coef_x in xx:
                dx = mx.node_number()
                step = self.zero()
                for pi in IntegerVectors(k - dx, n, min_part=vy, max_part=max_y):
                    for ly in product(*[y_homog[part] for part in pi]):
                        coef_y = basering.prod(mc[1] for mc in ly)
                        arbres_y = [mc[0] for mc in ly]
                        step += coef_y * self.sum(self(t)
                                                  for t in corolla_gen(mx, arbres_y, labels))
                resu += coef_x * step
        return resu

    def group_product(self, x, y, n, N=10):
        r"""
        Return the truncated group product of ``x`` and ``y``.

        This is a weighted sum of all corollas with up to ``n`` leaves, with
        ``x`` as root and ``y`` as leaves.

        The result is computed up to order ``N`` (included).

        When considered with infinitely many terms and infinite precision,
        this is an analogue of the Baker-Campbell-Hausdorff formula: it
        defines an associative product on the completed free pre-Lie algebra.

        INPUT:

        - ``x``, ``y`` -- two elements
        - ``n`` -- integer; the maximal width of corollas
        - ``N`` -- integer (default: 10); truncation order

        EXAMPLES:

        In the free pre-Lie algebra with one generator::

            sage: PL = algebras.FreePreLie(QQ)
            sage: a = PL.gen(0)
            sage: PL.group_product(a, a, 3, 3)
            B[[]] + B[[[]]] + 1/2*B[[[], []]]

        In the free pre-Lie algebra with several generators::

            sage: PL = algebras.FreePreLie(QQ,'@O')
            sage: a, b = PL.gens()
            sage: PL.group_product(a, b, 3, 3)
            B[@[]] + B[@[O[]]] + 1/2*B[@[O[], O[]]]
            sage: PL.group_product(a, b, 3, 10)
            B[@[]] + B[@[O[]]] + 1/2*B[@[O[], O[]]] + 1/6*B[@[O[], O[], O[]]]
        """
        br = self.base_ring()
        return x + self.sum(self.corolla(x, y, i, N) * ~br(factorial(i))
                            for i in range(1, n + 1))

    def _element_constructor_(self, x):
        r"""
        Convert ``x`` into ``self``.

        EXAMPLES::

            sage: R = algebras.FreePreLie(QQ, 'xy')
            sage: x, y = R.gens()
            sage: R(x)
            B[x[]]
            sage: R(x+4*y)
            B[x[]] + 4*B[y[]]

            sage: Trees = R.basis().keys()
            sage: R(Trees([],'x'))
            B[x[]]
            sage: D = algebras.FreePreLie(ZZ, 'xy')
            sage: X, Y = D.gens()
            sage: R(X-Y).parent()
            Free PreLie algebra on 2 generators ['x', 'y'] over Rational Field

        TESTS::

            sage: R.<x,y> = algebras.FreePreLie(QQ)
            sage: S.<z> = algebras.FreePreLie(GF(3))
            sage: R(z)
            Traceback (most recent call last):
            ...
            TypeError: not able to convert this to this algebra
        """
        if (isinstance(x, (RootedTree, LabelledRootedTree)) and
                x in self.basis().keys()):
            return self.monomial(x)
        try:
            P = x.parent()
            if isinstance(P, FreePreLieAlgebra):
                if P is self:
                    return x
                if self._coerce_map_from_(P):
                    return self.element_class(self, x.monomial_coefficients())
        except AttributeError:
            raise TypeError('not able to convert this to this algebra')
        else:
            raise TypeError('not able to convert this to this algebra')
        # Ok, not a pre-Lie algebra element (or should not be viewed as one).

    def _coerce_map_from_(self, R):
        r"""
        Return ``True`` if there is a coercion from ``R`` into ``self``
        and ``False`` otherwise.

        The things that coerce into ``self`` are

        - free pre-Lie algebras whose set `E` of labels is
          a subset of the corresponding ``self`` of ``set``, and whose base
          ring has a coercion map into ``self.base_ring()``

        EXAMPLES::

            sage: F = algebras.FreePreLie(GF(7), 'xyz'); F
            Free PreLie algebra on 3 generators ['x', 'y', 'z']
             over Finite Field of size 7

        Elements of the free pre-Lie algebra canonically coerce in::

            sage: x, y, z = F.gens()
            sage: F.coerce(x+y) == x+y
            True

        The free pre-Lie algebra over `\ZZ` on `x, y, z` coerces in, since
        `\ZZ` coerces to `\GF{7}`::

            sage: G = algebras.FreePreLie(ZZ, 'xyz')
            sage: Gx,Gy,Gz = G.gens()
            sage: z = F.coerce(Gx+Gy); z
            B[x[]] + B[y[]]
            sage: z.parent() is F
            True

        However, `\GF{7}` does not coerce to `\ZZ`, so the free pre-Lie
        algebra over `\GF{7}` does not coerce to the one over `\ZZ`::

            sage: G.coerce(y)
            Traceback (most recent call last):
            ...
            TypeError: no canonical coercion from Free PreLie algebra
             on 3 generators ['x', 'y', 'z'] over Finite Field of size
             7 to Free PreLie algebra on 3 generators ['x', 'y', 'z']
             over Integer Ring

        TESTS::

            sage: F = algebras.FreePreLie(ZZ, 'xyz')
            sage: G = algebras.FreePreLie(QQ, 'xyz')
            sage: H = algebras.FreePreLie(ZZ, 'y')
            sage: F._coerce_map_from_(G)
            False
            sage: G._coerce_map_from_(F)
            True
            sage: F._coerce_map_from_(H)
            True
            sage: F._coerce_map_from_(QQ)
            False
            sage: G._coerce_map_from_(QQ)
            False
            sage: F.has_coerce_map_from(PolynomialRing(ZZ, 3, 'x,y,z'))
            False
        """
        # free prelie algebras in a subset of variables
        # over any base that coerces in:
        if isinstance(R, FreePreLieAlgebra):
            if all(x in self.variable_names() for x in R.variable_names()):
                if self.base_ring().has_coerce_map_from(R.base_ring()):
                    return True
        return False

    def _construct_UEA(self):
        """
        Build the universal enveloping algebra.

        This is a Grossman-Larson Hopf algebra, based on forests of rooted
        trees.

        EXAMPLES::

            sage: S = algebras.FreePreLie(QQ, 'zt')
            sage: S._construct_UEA()
            Grossman-Larson Hopf algebra on 2 generators ['z', 't']
            over Rational Field
        """
        return GrossmanLarsonAlgebra(self.base_ring(), self.variable_names())

    def construction(self):
        """
        Return a pair ``(F, R)``, where ``F`` is a :class:`PreLieFunctor`
        and `R` is a ring, such that ``F(R)`` returns ``self``.

        EXAMPLES::

            sage: P = algebras.FreePreLie(ZZ, 'x,y')
            sage: x,y = P.gens()
            sage: F, R = P.construction()
            sage: F
            PreLie[x,y]
            sage: R
            Integer Ring
            sage: F(ZZ) is P
            True
            sage: F(QQ)
            Free PreLie algebra on 2 generators ['x', 'y'] over Rational Field
        """
        return PreLieFunctor(self.variable_names()), self.base_ring()

    class Element(CombinatorialFreeModule.Element):
        def lift(self):
            """
            Lift element to the Grossman-Larson algebra.

            EXAMPLES::

                sage: F = algebras.FreePreLie(QQ,'abc')
                sage: elt = F.an_element().lift(); elt
                B[#[a[a[a[a[]]]]]] + B[#[a[a[], a[a[]]]]]
                sage: parent(elt)
                Grossman-Larson Hopf algebra on 3 generators ['a', 'b', 'c']
                over Rational Field
            """
            UEA = self.parent()._construct_UEA()
            LRT = UEA.basis().keys()
            data = {LRT([x], ROOT): cf
                    for x, cf in self.monomial_coefficients(copy=False).items()}
            return UEA.element_class(UEA, data)

        def valuation(self):
            """
            Return the valuation of ``self``.

            EXAMPLES::

                sage: a = algebras.FreePreLie(QQ).gen(0)
                sage: a.valuation()
                1
                sage: (a*a).valuation()
                2

                sage: a, b = algebras.FreePreLie(QQ,'ab').gens()
                sage: (a+b).valuation()
                1
                sage: (a*b).valuation()
                2
                sage: (a*b+a).valuation()
                1

            TESTS::

                sage: z = algebras.FreePreLie(QQ).zero()
                sage: z.valuation()
                +Infinity
            """
            if self == self.parent().zero():
                return Infinity
            i = 0
            while True:
                i += 1
                if self.homogeneous_component(i):
                    return i


class PreLieFunctor(ConstructionFunctor):
    """
    A constructor for pre-Lie algebras.

    EXAMPLES::

        sage: P = algebras.FreePreLie(ZZ, 'x,y')
        sage: x,y = P.gens()
        sage: F = P.construction()[0]; F
        PreLie[x,y]

        sage: A = GF(5)['a,b']
        sage: a, b = A.gens()
        sage: F(A)
        Free PreLie algebra on 2 generators ['x', 'y'] over Multivariate Polynomial Ring in a, b over Finite Field of size 5

        sage: f = A.hom([a+b,a-b],A)
        sage: F(f)
        Generic endomorphism of Free PreLie algebra on 2 generators ['x', 'y']
        over Multivariate Polynomial Ring in a, b over Finite Field of size 5

        sage: F(f)(a * F(A)(x))
        (a+b)*B[x[]]
    """
    rank = 9

    def __init__(self, vars):
        """
        EXAMPLES::

            sage: F = sage.combinat.free_prelie_algebra.PreLieFunctor(['x','y'])
            sage: F
            PreLie[x,y]
            sage: F(ZZ)
            Free PreLie algebra on 2 generators ['x', 'y']  over Integer Ring
        """
        Functor.__init__(self, Rings(), Magmas())
        self.vars = vars

    def _apply_functor(self, R):
        """
        Apply the functor to an object of ``self``'s domain.

        EXAMPLES::

            sage: R = algebras.FreePreLie(ZZ, 'x,y,z')
            sage: F = R.construction()[0]; F
            PreLie[x,y,z]
            sage: type(F)
            <class 'sage.combinat.free_prelie_algebra.PreLieFunctor'>
            sage: F(ZZ)          # indirect doctest
            Free PreLie algebra on 3 generators ['x', 'y', 'z'] over Integer Ring
        """
        return FreePreLieAlgebra(R, self.vars)

    def _apply_functor_to_morphism(self, f):
        """
        Apply the functor ``self`` to the ring morphism `f`.

        TESTS::

            sage: R = algebras.FreePreLie(ZZ, 'x').construction()[0]
            sage: R(ZZ.hom(GF(3)))  # indirect doctest
            Generic morphism:
              From: Free PreLie algebra on one generator ['x'] over Integer Ring
              To:   Free PreLie algebra on one generator ['x'] over Finite Field of size 3
        """
        dom = self(f.domain())
        codom = self(f.codomain())

        def action(x):
            return codom._from_dict({a: f(b)
                                     for a, b in x.monomial_coefficients().items()})
        return dom.module_morphism(function=action, codomain=codom)

    def __eq__(self, other):
        """
        EXAMPLES::

            sage: F = algebras.FreePreLie(ZZ, 'x,y,z').construction()[0]
            sage: G = algebras.FreePreLie(QQ, 'x,y,z').construction()[0]
            sage: F == G
            True
            sage: G == loads(dumps(G))
            True
            sage: G = algebras.FreePreLie(QQ, 'x,y').construction()[0]
            sage: F == G
            False
        """
        if not isinstance(other, PreLieFunctor):
            return False
        return self.vars == other.vars

    def __mul__(self, other):
        """
        If two PreLie functors are given in a row, form a single PreLie functor
        with all of the variables.

        EXAMPLES::

            sage: F = sage.combinat.free_prelie_algebra.PreLieFunctor(['x','y'])
            sage: G = sage.combinat.free_prelie_algebra.PreLieFunctor(['t'])
            sage: G * F
            PreLie[x,y,t]
        """
        if isinstance(other, IdentityConstructionFunctor):
            return self
        if isinstance(other, PreLieFunctor):
            if set(self.vars).intersection(other.vars):
                raise CoercionException("Overlapping variables (%s,%s)" %
                                        (self.vars, other.vars))
            return PreLieFunctor(other.vars + self.vars)
        elif (isinstance(other, CompositeConstructionFunctor) and
              isinstance(other.all[-1], PreLieFunctor)):
            return CompositeConstructionFunctor(other.all[:-1],
                                                self * other.all[-1])
        else:
            return CompositeConstructionFunctor(other, self)

    def merge(self, other):
        """
        Merge ``self`` with another construction functor, or return None.

        EXAMPLES::

            sage: F = sage.combinat.free_prelie_algebra.PreLieFunctor(['x','y'])
            sage: G = sage.combinat.free_prelie_algebra.PreLieFunctor(['t'])
            sage: F.merge(G)
            PreLie[x,y,t]
            sage: F.merge(F)
            PreLie[x,y]

        Now some actual use cases::

            sage: R = algebras.FreePreLie(ZZ, 'xyz')
            sage: x,y,z = R.gens()
            sage: 1/2 * x
            1/2*B[x[]]
            sage: parent(1/2 * x)
            Free PreLie algebra on 3 generators ['x', 'y', 'z'] over Rational Field

            sage: S = algebras.FreePreLie(QQ, 'zt')
            sage: z,t = S.gens()
            sage: x + t
            B[t[]] + B[x[]]
            sage: parent(x + t)
            Free PreLie algebra on 4 generators ['z', 't', 'x', 'y'] over Rational Field
        """
        if isinstance(other, PreLieFunctor):
            if self.vars == other.vars:
                return self
            ret = list(self.vars)
            cur_vars = set(ret)
            ret.extend(v for v in other.vars if v not in cur_vars)
            return PreLieFunctor(Alphabet(ret))

        return None

    def _repr_(self) -> str:
        """
        TESTS::

            sage: algebras.FreePreLie(QQ,'x,y,z,t').construction()[0]
            PreLie[x,y,z,t]
        """
        return "PreLie[%s]" % ','.join(self.vars)


def tree_from_sortkey(ch, labels=True):
    r"""
    Transform a list of ``(valence, label)`` into a tree and a remainder.

    This is like an inverse of the ``sort_key`` method.

    INPUT:

    - ``ch`` -- list of pairs ``(integer, label)``
    - ``labels`` -- boolean (default: ``True``); whether to use labelled trees

    OUTPUT:

    a pair ``(tree, remainder of the input)``

    EXAMPLES::

        sage: from sage.combinat.free_prelie_algebra import tree_from_sortkey
        sage: a = algebras.FreePreLie(QQ).gen(0)
        sage: t = (a*a*a*a).support()
        sage: all(tree_from_sortkey(u.sort_key(), False)[0] == u for u in t)
        True

        sage: a, b = algebras.FreePreLie(QQ,'ab').gens()
        sage: t = (a*b*a*b).support()
        sage: all(tree_from_sortkey(u.sort_key())[0] == u for u in t)
        True
    """
    if labels:
        Trees = LabelledRootedTrees()
        width, label = ch[0]
    else:
        Trees = RootedTrees()
        width = ch[0]

    remainder = ch[1:]
    if width == 0:
        if labels:
            return (Trees([], label), remainder)
        return (Trees([]), remainder)

    branches = {}
    for i in range(width):
        tree, remainder = tree_from_sortkey(remainder, labels=labels)
        branches[i] = tree

    if labels:
        return (Trees(branches.values(), label), remainder)
    return (Trees(branches.values()), remainder)


def corolla_gen(tx, list_ty, labels=True):
    """
    Yield the terms in the corolla with given bottom tree and top trees.

    These are the possible terms in the simultaneous grafting of the
    top trees on vertices of the bottom tree.

    INPUT:

    - ``tx`` -- a tree
    - ``list_ty`` -- list of trees

    EXAMPLES::

        sage: from sage.combinat.free_prelie_algebra import corolla_gen
        sage: a = algebras.FreePreLie(QQ).gen(0)
        sage: ta = a.support()[0]
        sage: list(corolla_gen(ta,[ta],False))
        [[[]]]

        sage: a, b = algebras.FreePreLie(QQ,'ab').gens()
        sage: ta = a.support()[0]
        sage: tb = b.support()[0]
        sage: ab = (a*b).support()[0]
        sage: list(corolla_gen(ta,[tb]))
        [a[b[]]]
        sage: list(corolla_gen(tb,[ta,ta]))
        [b[a[], a[]]]
        sage: list(corolla_gen(ab,[ab,ta]))
        [a[a[], b[], a[b[]]], a[a[b[]], b[a[]]], a[a[], b[a[b[]]]],
        a[b[a[], a[b[]]]]]
    """
    n = len(list_ty)
    zx = tx.sort_key()
    nx = len(zx)
    liste_zy = [t.sort_key() for t in list_ty]
    for list_pos in product(range(nx), repeat=n):
        new_zx = tuple(zx)
        data = zip(list_pos, liste_zy)
        sorted_data = sorted(data, reverse=True)
        for pos_t in sorted_data:
            if labels:
                idx, lbl = new_zx[pos_t[0]]
                new_zx = (new_zx[:pos_t[0]] + ((idx + 1, lbl),) +
                          pos_t[1] + new_zx[pos_t[0] + 1:])
            else:
                idx = new_zx[pos_t[0]]
                new_zx = (new_zx[:pos_t[0]] + (idx + 1,) +
                          pos_t[1] + new_zx[pos_t[0] + 1:])
        yield tree_from_sortkey(new_zx, labels=labels)[0]



# sage_setup: distribution = sagemath-combinat
r"""
Symmetric functions in non-commuting variables

- :class:`Introduction to Symmetric Functions in Non-Commuting Variables <sage.combinat.ncsym.ncsym.SymmetricFunctionsNonCommutingVariables>`

- :ref:`sage.combinat.ncsym.bases`
- :ref:`sage.combinat.ncsym.dual`
- :ref:`sage.combinat.ncsym.ncsym`
"""
# install the docstring of this module to the containing package
from sage.misc.namespace_package import install_doc
install_doc(__package__, __doc__)

from sage.misc.lazy_import import lazy_import

lazy_import('sage.combinat.ncsym.ncsym', 'SymmetricFunctionsNonCommutingVariables')
lazy_import('sage.combinat.ncsym.dual', 'SymmetricFunctionsNonCommutingVariablesDual')

del install_doc
del lazy_import



# sage_setup: distribution = sagemath-combinat
"""
Graftal Lace Cellular Automata

AUTHORS:

- Travis Scrimshaw (2020-04-30): Initial version
"""

#*****************************************************************************
#       Copyright (C) 2020 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.sage_object import SageObject
from sage.typeset.ascii_art import AsciiArt
from sage.typeset.unicode_art import UnicodeArt


class GraftalLaceCellularAutomata(SageObject):
    r"""
    Graftal Lace Cellular Automata (GLCA).

    A GLCA is a deterministic cellular automaton whose rule is given
    by an 8-digit octal number `r_7 \cdots r_0`. For a node `s_i`, let
    `b_k`, for `k = -1,0,1` denote if there is an edge from `s_i` to
    `s'_{i+k}`, where `s'_j` is the previous row. We determine the value
    at `t_{i+k}` by considering the value of `r_m`, where the binary
    representation of `m` is `b_{-1} b_0 b_1`. If `r_m` has a binary
    representation of b'_1 b'_0 b'_{-1}`, then we add `b'_k` to `t_{i+k}`.

    INPUT:

    - ``rule`` -- list of length 8 with integer entries `0 \leq x < 8`

    EXAMPLES::

        sage: G = cellular_automata.GraftalLace([0,2,5,4,7,2,3,3])
        sage: G.evolve(3)
        sage: ascii_art(G)
               o
               |
               o
               |
             o o o
            /| |/|
           o o o o o
          /| |/|\|/|
         o o o o o o o

        sage: G = cellular_automata.GraftalLace([3,0,3,4,7,6,3,1])
        sage: G.evolve(3)
        sage: ascii_art(G)
               o
               |
               o
               |\
             o o o
            /  |\ \
           o o o o o
          /|/  |\ \ \
         o o o o o o o

        sage: G = cellular_automata.GraftalLace([2,0,3,3,6,0,2,7])
        sage: G.evolve(20)
        sage: G.plot()                                                                  # needs sage.plot
        Graphics object consisting of 842 graphics primitives

    .. PLOT::

        G = cellular_automata.GraftalLace([2,0,3,3,6,0,2,7])
        G.evolve(20)
        P = G.plot()
        sphinx_plot(P)

    REFERENCES:

    - [Kas2018]_
    """
    def __init__(self, rule):
        """
        Initialize ``self``.

        TESTS::

            sage: G = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: TestSuite(G).run()

            sage: G = cellular_automata.GraftalLace([5,1,2,5,4,5,5])
            Traceback (most recent call last):
            ...
            ValueError: invalid rule
            sage: G = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0,5])
            Traceback (most recent call last):
            ...
            ValueError: invalid rule
            sage: G = cellular_automata.GraftalLace([5,1,2,5,-1,5,5,0])
            Traceback (most recent call last):
            ...
            ValueError: invalid rule
            sage: G = cellular_automata.GraftalLace([8,5,1,2,5,4,5,5])
            Traceback (most recent call last):
            ...
            ValueError: invalid rule
        """
        if len(rule) != 8 or any(x not in range(8) for x in rule):
            raise ValueError("invalid rule")
        # We reverse the rule to make the evolution easier to compute
        self._rule = tuple(reversed(rule))
        self._states = [[2]]

    def __eq__(self, other):
        """
        Check equality.

        Two GLCAs are equal if and only if they have the same rule.

        TESTS::

            sage: G1 = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: G2 = cellular_automata.GraftalLace([0,5,5,4,5,2,1,5])
            sage: G3 = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: G1 == G2
            False
            sage: G1 == G3
            True
            sage: G1 is G3
            False
        """
        return (isinstance(other, GraftalLaceCellularAutomata)
                and self._rule == other._rule)

    def __ne__(self, other):
        """
        Check non equality.

        TESTS::

            sage: G1 = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: G2 = cellular_automata.GraftalLace([0,5,5,4,5,2,1,5])
            sage: G3 = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: G1 != G2
            True
            sage: G1 != G3
            False
            sage: G1 is G3
            False
        """
        return not (self == other)

    # Evolution functions
    # -------------------

    def evolve(self, number=None):
        r"""
        Evolve ``self``.

        INPUT:

        - ``number`` -- (default: 1) the number of times to perform
          the evolution

        EXAMPLES::

            sage: G = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: ascii_art(G)
             o
             |
             o
            sage: G.evolve(2)
            sage: ascii_art(G)
                 o
                 |
                 o
                / \
               o o o
              / \ / \
             o o o o o

            sage: G = cellular_automata.GraftalLace([0,2,1,4,7,2,3,0])
            sage: G.evolve(3)
            sage: ascii_art(G)
                   o
                   |
                   o
                   |
                 o o o
                   |
               o o o o o
                   |
             o o o o o o o
        """
        if number is not None:
            for _ in range(number):
                self.evolve()
            return

        prev_state = self._states[-1]
        next_state = [0] * (len(prev_state) + 2)

        for i, val in enumerate(prev_state):
            next_state[i] += self._rule[val] & 0x1
            next_state[i+1] += self._rule[val] & 0x2
            next_state[i+2] += self._rule[val] & 0x4
        self._states.append(next_state)

    # Output functions
    # ----------------

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            Graftal Lace Cellular Automata with rule 51254550
        """
        rule = ''.join(map(str, reversed(self._rule)))
        return "Graftal Lace Cellular Automata with rule {}".format(rule)

    def print_states(self, number=None, use_unicode=False):
        r"""
        Print the first ``num`` states of ``self``.

        .. NOTE::

            If the number of states computed for ``self`` is less than
            ``num``, then this evolves the system using the default
            time evolution.

        INPUT:

        - ``number`` -- the number of states to print

        EXAMPLES::

            sage: G = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: G.evolve(2)
            sage: G.print_states()
                 o
                 |
                 o
                / \
               o o o
              / \ / \
             o o o o o
            sage: G.evolve(20)
            sage: G.print_states(3)
                 o
                 |
                 o
                / \
               o o o
              / \ / \
             o o o o o
        """
        if number is None:
            number = len(self._states)
        if number > len(self._states):
            for dummy in range(number - len(self._states)):
                self.evolve()
        if use_unicode:
            print(self._unicode_art_(number))
        else:
            print(self._ascii_art_(number))

    def _ascii_art_(self, number=None):
        r"""
        Return an ascii art representation of ``self``.

        EXAMPLES::

            sage: G = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: G.evolve(10)
            sage: ascii_art(G)
                                 o
                                 |
                                 o
                                / \
                               o o o
                              / \ / \
                             o o o o o
                            / \  |  / \
                           o o o o o o o
                          / \ / X X \ / \
                         o o o o o o o o o
                        / \  |/ \|/ \|  / \
                       o o o o o o o o o o o
                      / \ / \ \ / \ / / \ / \
                     o o o o o o o o o o o o o
                    / \  |  / \|   |/ \  |  / \
                   o o o o o o o o o o o o o o o
                  / \ / X X \ /     \ / X X \ / \
                 o o o o o o o o o o o o o o o o o
                / \  |/ \|/ \|       |/ \|/ \|  / \
               o o o o o o o o o o o o o o o o o o o
              / \ / \ \ / \ /         \ / \ / / \ / \
             o o o o o o o o o o o o o o o o o o o o o
        """
        if number is None:
            number = len(self._states)

        space = len(self._states[:number]) * 2 - 1
        ret = AsciiArt([' '*space + 'o'])
        space += 1
        for i,state in enumerate(self._states[:number]):
            temp = ' '*(space-2)
            last = ' '
            for x in state:
                if x & 0x4:
                    if last == '/':
                        temp += 'X'
                    else:
                        temp += '\\'
                else:
                    temp += last
                temp += '|' if x & 0x2 else ' '
                last = '/' if x & 0x1 else ' '
            ret *= AsciiArt([temp + last])
            space -= 1
            ret *= AsciiArt([' '*space + ' '.join('o' for dummy in range(2*i+1))])
            space -= 1
        return ret

    def _unicode_art_(self, number=None):
        r"""
        Return a unicode art representation of ``self``.

        EXAMPLES::

            sage: G = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: G.evolve(10)
            sage: unicode_art(G)
                                 ◾
                                 │
                                 ◾
                                ╱ ╲
                               ◾ ◾ ◾
                              ╱ ╲ ╱ ╲
                             ◾ ◾ ◾ ◾ ◾
                            ╱ ╲  │  ╱ ╲
                           ◾ ◾ ◾ ◾ ◾ ◾ ◾
                          ╱ ╲ ╱ ╳ ╳ ╲ ╱ ╲
                         ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾
                        ╱ ╲  │╱ ╲│╱ ╲│  ╱ ╲
                       ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾
                      ╱ ╲ ╱ ╲ ╲ ╱ ╲ ╱ ╱ ╲ ╱ ╲
                     ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾
                    ╱ ╲  │  ╱ ╲│   │╱ ╲  │  ╱ ╲
                   ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾
                  ╱ ╲ ╱ ╳ ╳ ╲ ╱     ╲ ╱ ╳ ╳ ╲ ╱ ╲
                 ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾
                ╱ ╲  │╱ ╲│╱ ╲│       │╱ ╲│╱ ╲│  ╱ ╲
               ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾
              ╱ ╲ ╱ ╲ ╲ ╱ ╲ ╱         ╲ ╱ ╲ ╱ ╱ ╲ ╱ ╲
             ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾ ◾
        """
        if number is None:
            number = len(self._states)

        space = len(self._states[:number]) * 2 - 1
        ret = UnicodeArt([' '*space + '◾'])
        space += 1
        for i,state in enumerate(self._states[:number]):
            temp = ' '*(space-2)
            last = ' '
            for x in state:
                if x & 0x4:
                    if last == '╱':
                        temp += '╳'
                    else:
                        temp += '╲'
                else:
                    temp += last
                temp += '│' if x & 0x2 else ' '
                last = '╱' if x & 0x1 else ' '
            ret *= UnicodeArt([temp + last])
            space -= 1
            ret *= UnicodeArt([' '*space + ' '.join('◾' for dummy in range(2*i+1))])
            space -= 1
        return ret

    def plot(self, number=None):
        r"""
        Return a plot of ``self``.

        INPUT:

        - ``number`` -- the number of states to plot

        EXAMPLES::

            sage: G = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: G.evolve(20)
            sage: G.plot()                                                              # needs sage.plot
            Graphics object consisting of 865 graphics primitives
        """
        if number is None:
            number = len(self._states)
        if number > len(self._states):
            for dummy in range(number - len(self._states)):
                self.evolve()

        from sage.plot.circle import circle
        from sage.plot.line import line

        x = len(self._states[:number])
        rad = 0.1
        ret = circle((x, 1), rad, fill=True)
        for i,state in enumerate(self._states[:number]):
            for j,val in enumerate(state):
                if val & 0x4:
                    ret += line([(x+j,-i), (x+j-1,-i+1)])
                if val & 0x2:
                    ret += line([(x+j,-i), (x+j,-i+1)])
                if val & 0x1:
                    ret += line([(x+j,-i), (x+j+1,-i+1)])
            for j in range(2*i+1):
                ret += circle((x+j, -i), rad, fill=True)
            x -= 1
        ret.set_aspect_ratio(1)
        ret.axes(False)
        return ret

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: G = cellular_automata.GraftalLace([5,1,2,5,4,5,5,0])
            sage: G.evolve(2)
            sage: latex(G)
            \begin{tikzpicture}
            \fill (3,1) circle (2pt);
            \draw[-] (3,0) -- (3,1);
            \fill (3,0) circle (2pt);
            \draw[-] (2,-1) -- (3,0);
            \draw[-] (4,-1) -- (3,0);
            \fill (2,-1) circle (2pt);
            \fill (3,-1) circle (2pt);
            \fill (4,-1) circle (2pt);
            \draw[-] (1,-2) -- (2,-1);
            \draw[-] (3,-2) -- (2,-1);
            \draw[-] (3,-2) -- (4,-1);
            \draw[-] (5,-2) -- (4,-1);
            \fill (1,-2) circle (2pt);
            \fill (2,-2) circle (2pt);
            \fill (3,-2) circle (2pt);
            \fill (4,-2) circle (2pt);
            \fill (5,-2) circle (2pt);
            \end{tikzpicture}
        """
        ret = "\\begin{tikzpicture}\n"

        x = len(self._states)
        rad = 2
        ret += "\\fill ({},{}) circle ({}pt);\n".format(x, 1, rad)
        for i,state in enumerate(self._states):
            for j,val in enumerate(state):
                if val & 0x4:
                    ret += "\\draw[-] ({},{}) -- ({},{});\n".format(x+j,-i, x+j-1,-i+1)
                if val & 0x2:
                    ret += "\\draw[-] ({},{}) -- ({},{});\n".format(x+j,-i, x+j,-i+1)
                if val & 0x1:
                    ret += "\\draw[-] ({},{}) -- ({},{});\n".format(x+j,-i, x+j+1,-i+1)
            for j in range(2*i+1):
                ret += "\\fill ({},{}) circle ({}pt);\n".format(x+j, -i, rad)
            x -= 1
        ret += "\\end{tikzpicture}"
        return ret



# sage_setup: distribution = sagemath-combinat
r"""
Crystals

- :ref:`sage.combinat.crystals.crystals`
- The `Lie Methods and Related Combinatorics <../../../../../thematic_tutorials/lie.html>`_ thematic tutorial
- :ref:`sage.combinat.crystals.catalog`

See also the categories for crystals: :class:`Crystals`,
:class:`HighestWeightCrystals`, :class:`FiniteCrystals`,
:class:`ClassicalCrystals`, :class:`RegularCrystals`,
:class:`~sage.categories.regular_supercrystals.RegularSuperCrystals`
"""
# install the docstring of this module to the containing package
from sage.misc.namespace_package import install_doc
install_doc(__package__, __doc__)

from sage.misc.lazy_import import lazy_import

lazy_import('sage.combinat.crystals', 'catalog', 'crystals')
del lazy_import
del install_doc



# sage_setup: distribution = sagemath-combinat
r"""
Six vertex model
"""

from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.list_clone import ClonableArray
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.combinat.combinatorial_map import combinatorial_map


class SixVertexConfiguration(ClonableArray):
    """
    A configuration in the six vertex model.
    """

    def check(self):
        """
        Check if ``self`` is a valid 6 vertex configuration.

        EXAMPLES::

            sage: M = SixVertexModel(3, boundary_conditions='ice')
            sage: M[0].check()
        """
        if self not in self.parent():
            raise ValueError("invalid configuration")

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: M = SixVertexModel(3, boundary_conditions='ice')
            sage: M[0]
                ^    ^    ^
                |    |    |
            --> # <- # <- # <--
                |    ^    ^
                V    |    |
            --> # -> # <- # <--
                |    |    ^
                V    V    |
            --> # -> # -> # <--
                |    |    |
                V    V    V
        """
        # List are in the order of URDL
        ascii = [[r'  V  ', ' -', r'  ^  ', '- '], # LR
                 [r'  |  ', ' <', r'  ^  ', '- '], # LU
                 [r'  V  ', ' <', r'  |  ', '- '], # LD
                 [r'  |  ', ' <', r'  |  ', '> '], # UD
                 [r'  |  ', ' -', r'  ^  ', '> '], # UR
                 [r'  V  ', ' -', r'  |  ', '> ']] # RD
        ret = '  '
        # Do the top line
        for entry in self[0]:
            if entry == 1 or entry == 3 or entry == 4:
                ret += '  ^  '
            else:
                ret += '  |  '

        # Do the meat of the ascii art
        for row in self:
            ret += '\n  '
            # Do the top row
            for entry in row:
                ret += ascii[entry][0]
            ret += '\n'

            # Do the left-most entry
            if row[0] == 0 or row[0] == 1 or row[0] == 2:
                ret += '<-'
            else:
                ret += '--'

            # Do the middle row
            for entry in row:
                ret += ascii[entry][3] + '#' + ascii[entry][1]

            # Do the right-most entry
            if row[-1] == 0 or row[-1] == 4 or row[-1] == 5:
                ret += '->'
            else:
                ret += '--'

            # Do the bottom row
            ret += '\n  '
            for entry in row:
                ret += ascii[entry][2]

        # Do the bottom line
        ret += '\n  '
        for entry in self[-1]:
            if entry == 2 or entry == 3 or entry == 5:
                ret += '  V  '
            else:
                ret += '  |  '

        return ret

    def to_signed_matrix(self):
        """
        Return the signed matrix of ``self``.

        The signed matrix corresponding to a six vertex configuration is
        given by `0` if there is a cross flow, a `1` if the outward arrows
        are vertical and `-1` if the outward arrows are horizontal.

        EXAMPLES::

            sage: M = SixVertexModel(3, boundary_conditions='ice')
            sage: [x.to_signed_matrix() for x in M]                                     # needs sage.modules
            [
            [1 0 0]  [1 0 0]  [ 0  1  0]  [0 1 0]  [0 1 0]  [0 0 1]  [0 0 1]
            [0 1 0]  [0 0 1]  [ 1 -1  1]  [1 0 0]  [0 0 1]  [1 0 0]  [0 1 0]
            [0 0 1], [0 1 0], [ 0  1  0], [0 0 1], [1 0 0], [0 1 0], [1 0 0]
            ]
        """
        from sage.matrix.constructor import matrix
        # verts = ['LR', 'LU', 'LD', 'UD', 'UR', 'RD']

        def matrix_sign(x):
            if x == 0:
                return -1
            if x == 3:
                return 1
            return 0
        return matrix([[matrix_sign(r) for r in row] for row in self])

    def plot(self, color='sign'):
        """
        Return a plot of ``self``.

        INPUT:

        - ``color`` -- can be any of the following:

          * ``4`` -- use 4 colors: black, red, blue, and green with each
            corresponding to up, right, down, and left respectively
          * ``2`` -- use 2 colors: red for horizontal, blue for vertical arrows
          * ``'sign'`` -- use red for right and down arrows, blue for left
            and up arrows
          * a list of 4 colors for each direction
          * a function which takes a direction and a boolean corresponding
            to the sign

        EXAMPLES::

            sage: M = SixVertexModel(2, boundary_conditions='ice')
            sage: print(M[0].plot().description())                                      # needs sage.plot
            Arrow from (-1.0,0.0) to (0.0,0.0)
            Arrow from (-1.0,1.0) to (0.0,1.0)
            Arrow from (0.0,0.0) to (0.0,-1.0)
            Arrow from (0.0,0.0) to (1.0,0.0)
            Arrow from (0.0,1.0) to (0.0,0.0)
            Arrow from (0.0,1.0) to (0.0,2.0)
            Arrow from (1.0,0.0) to (1.0,-1.0)
            Arrow from (1.0,0.0) to (1.0,1.0)
            Arrow from (1.0,1.0) to (0.0,1.0)
            Arrow from (1.0,1.0) to (1.0,2.0)
            Arrow from (2.0,0.0) to (1.0,0.0)
            Arrow from (2.0,1.0) to (1.0,1.0)
        """
        from sage.plot.graphics import Graphics
        from sage.plot.arrow import arrow

        if color == 4:
            color_list = ['black', 'red', 'blue', 'green']
            cfunc = lambda d,pm: color_list[d]
        elif color == 2:
            cfunc = lambda d,pm: 'red' if d % 2 == 0 else 'blue'
        elif color == 1 or color is None:
            cfunc = lambda d,pm: 'black'
        elif color == 'sign':
            cfunc = lambda d,pm: 'red' if pm else 'blue' # RD are True
        elif isinstance(color, (list, tuple)):
            cfunc = lambda d,pm: color[d]
        else:
            cfunc = color

        G = Graphics()
        for j,row in enumerate(reversed(self)):
            for i,entry in enumerate(row):
                if entry == 0: # LR
                    G += arrow((i,j+1), (i,j), color=cfunc(2, True))
                    G += arrow((i,j), (i+1,j), color=cfunc(1, True))
                    if j == 0:
                        G += arrow((i,j-1), (i,j), color=cfunc(0, False))
                    if i == 0:
                        G += arrow((i,j), (i-1,j), color=cfunc(3, False))
                elif entry == 1: # LU
                    G += arrow((i,j), (i,j+1), color=cfunc(0, False))
                    G += arrow((i+1,j), (i,j), color=cfunc(3, False))
                    if j == 0:
                        G += arrow((i,j-1), (i,j), color=cfunc(0, False))
                    if i == 0:
                        G += arrow((i,j), (i-1,j), color=cfunc(3, False))
                elif entry == 2: # LD
                    G += arrow((i,j+1), (i,j), color=cfunc(2, True))
                    G += arrow((i+1,j), (i,j), color=cfunc(3, False))
                    if j == 0:
                        G += arrow((i,j), (i,j-1), color=cfunc(2, True))
                    if i == 0:
                        G += arrow((i,j), (i-1,j), color=cfunc(3, False))
                elif entry == 3: # UD
                    G += arrow((i,j), (i,j+1), color=cfunc(0, False))
                    G += arrow((i+1,j), (i,j), color=cfunc(3, False))
                    if j == 0:
                        G += arrow((i,j), (i,j-1), color=cfunc(2, True))
                    if i == 0:
                        G += arrow((i-1,j), (i,j), color=cfunc(1, True))
                elif entry == 4: # UR
                    G += arrow((i,j), (i,j+1), color=cfunc(0, False))
                    G += arrow((i,j), (i+1,j), color=cfunc(1, True))
                    if j == 0:
                        G += arrow((i,j-1), (i,j), color=cfunc(0, False))
                    if i == 0:
                        G += arrow((i-1,j), (i,j), color=cfunc(1, True))
                elif entry == 5: # RD
                    G += arrow((i,j+1), (i,j), color=cfunc(2, True))
                    G += arrow((i,j), (i+1,j), color=cfunc(1, True))
                    if j == 0:
                        G += arrow((i,j), (i,j-1), color=cfunc(2, True))
                    if i == 0:
                        G += arrow((i-1,j), (i,j), color=cfunc(1, True))
        G.axes(False)
        return G

    def energy(self, epsilon):
        r"""
        Return the energy of the configuration.

        The energy of a configuration `\nu` is defined as

        .. MATH::

            E(\nu) = n_0 \epsilon_0 + n_1 \epsilon_1 + \cdots + n_5 \epsilon_5

        where `n_i` is the number of vertices of type `i` and
        `\epsilon_i` is the `i`-th energy constant.

        .. NOTE::

            We number our configurations as:

            0. LR
            1. LU
            2. LD
            3. UD
            4. UR
            5. RD

            which differs from :wikipedia:`Ice-type_model`.

        EXAMPLES::

            sage: M = SixVertexModel(3, boundary_conditions='ice')
            sage: nu = M[2]; nu
                ^    ^    ^
                |    |    |
            --> # -> # <- # <--
                ^    |    ^
                |    V    |
            --> # <- # -> # <--
                |    ^    |
                V    |    V
            --> # -> # <- # <--
                |    |    |
                V    V    V
            sage: nu.energy([1,2,1,2,1,2])
            15

        A KDP energy::

            sage: nu.energy([1,1,0,1,0,1])
            7

        A Rys `F` energy::

            sage: nu.energy([0,1,1,0,1,1])
            4

        The zero field assumption::

            sage: nu.energy([1,2,3,1,3,2])
            15
        """
        if len(epsilon) != 6:
            raise ValueError("there must be 6 energy constants")
        return sum(epsilon[entry] for row in self for entry in row)


class SixVertexModel(UniqueRepresentation, Parent):
    """
    The six vertex model.

    We model a configuration by indicating which configuration by the
    following six configurations which are determined by the two outgoing
    arrows in the Up, Right, Down, Left directions:

    1. LR::

            |
            V
        <-- # -->
            ^
            |

    2. LU::

            ^
            |
        <-- # <--
            ^
            |

    3. LD::

            |
            V
        <-- # <--
            |
            V

    4. UD::

            ^
            |
        --> # <--
            |
            V

    5. UR::

            ^
            |
        --> # -->
            ^
            |
    6. RD::

            |
            V
        --> # -->
            |
            V

    INPUT:

    - ``n`` -- the number of rows
    - ``m`` -- (optional) the number of columns, if not specified, then
      the number of columns is the number of rows
    - ``boundary_conditions`` -- (optional) a quadruple of tuples whose
      entries are either:

      * ``True`` for an inward arrow,
      * ``False`` for an outward arrow, or
      * ``None`` for no boundary condition.

      There are also the following predefined boundary conditions:

      * ``'ice'`` -- the top and bottom boundary conditions are outward and the
        left and right boundary conditions are inward; this gives the square
        ice model. Also called domain wall boundary conditions.
      * ``'domain wall'`` -- same as ``'ice'``.
      * ``'alternating'`` -- the boundary conditions alternate between inward
        and outward.
      * ``'free'`` -- there are no boundary conditions.

    EXAMPLES:

    Here are the six types of vertices that can be created::

        sage: M = SixVertexModel(1)
        sage: list(M)
        [
            |          ^          |          ^          ^          |
            V          |          V          |          |          V
        <-- # -->  <-- # <--  <-- # <--  --> # <--  --> # -->  --> # -->
            ^          ^          |          |          ^          |
            |    ,     |    ,     V    ,     V    ,     |    ,     V
        ]

    When using the square ice model, it is known that the number of
    configurations is equal to the number of alternating sign matrices::

        sage: M = SixVertexModel(1, boundary_conditions='ice')
        sage: len(M)
        1
        sage: M = SixVertexModel(4, boundary_conditions='ice')
        sage: len(M)
        42
        sage: all(len(SixVertexModel(n, boundary_conditions='ice'))                     # needs sage.modules
        ....:     == AlternatingSignMatrices(n).cardinality() for n in range(1, 7))
        True

    An example with a specified non-standard boundary condition and
    non-rectangular shape::

        sage: M = SixVertexModel(2, 1, [[None], [True,True], [None], [None,None]])
        sage: list(M)
        [
            ^          ^          |          ^
            |          |          V          |
        <-- # <--  <-- # <--  <-- # <--  --> # <--
            ^          ^          |          |
            |          |          V          V
        <-- # <--  --> # <--  <-- # <--  <-- # <--
            ^          |          |          |
            |    ,     V    ,     V    ,     V
        ]

    REFERENCES:

    - :wikipedia:`Vertex_model`
    - :wikipedia:`Ice-type_model`
    """
    @staticmethod
    def __classcall_private__(cls, n, m=None, boundary_conditions=None):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: M1 = SixVertexModel(1, boundary_conditions=[[False],[True],[False],[True]])
            sage: M2 = SixVertexModel(1, 1, ((False,),(True,),(False,),(True,)))
            sage: M1 is M2
            True
        """
        if m is None:
            m = n
        if boundary_conditions is None or boundary_conditions == 'free':
            boundary_conditions = ((None,)*m, (None,)*n)*2
        elif boundary_conditions == 'alternating':
            bdry = True
            cond = []
            for dummy in range(2):
                val = []
                for _ in range(m):
                    val.append(bdry)
                    bdry = not bdry
                cond.append(tuple(val))
                val = []
                for k in range(n):
                    val.append(bdry)
                    bdry = not bdry
                cond.append(tuple(val))
            boundary_conditions = tuple(cond)
        elif boundary_conditions == 'ice' or boundary_conditions == 'domain wall':
            if m == n:
                return SquareIceModel(n)
            boundary_conditions = ((False,)*m, (True,)*n)*2
        else:
            boundary_conditions = tuple(tuple(x) for x in boundary_conditions)
        return super().__classcall__(cls, n, m, boundary_conditions)

    def __init__(self, n, m, boundary_conditions):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: M = SixVertexModel(2, boundary_conditions='ice')
            sage: TestSuite(M).run()
        """
        self._nrows = n
        self._ncols = m
        self._bdry_cond = boundary_conditions # Ordered URDL
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: SixVertexModel(2, boundary_conditions='ice')
            The six vertex model on a 2 by 2 grid
        """
        return "The six vertex model on a {} by {} grid".format(self._nrows, self._ncols)

    def _repr_option(self, key):
        """
        Metadata about the ``_repr_()`` output.

        See :meth:`sage.structure.parent._repr_option` for details.

        EXAMPLES::

            sage: M = SixVertexModel(2, boundary_conditions='ice')
            sage: M._repr_option('element_ascii_art')
            True
        """
        if key == 'element_ascii_art':
            return True
        return Parent._repr_option(self, key)

    def _element_constructor_(self, x):
        """
        Construct an element of ``self``.

        EXAMPLES::

            sage: M = SixVertexModel(2, boundary_conditions='ice')
            sage: M([[3,1],[5,3]])
                ^    ^
                |    |
            --> # <- # <--
                |    ^
                V    |
            --> # -> # <--
                |    |
                V    V
        """
        if isinstance(x, SixVertexConfiguration):
            if x.parent() is not self:
                return self.element_class(self, tuple(x))
            return x

        verts = ['LR', 'LU', 'LD', 'UD', 'UR', 'RD']
        elt = []
        for row in x:
            elt.append([])
            for entry in row:
                if entry in verts:
                    elt[-1].append(verts.index(entry))
                elif entry in range(6):
                    elt[-1].append(entry)
                else:
                    raise ValueError("invalid entry")
            elt[-1] = tuple(elt[-1])
        return self.element_class(self, tuple(elt))

    Element = SixVertexConfiguration

    def __iter__(self):
        """
        Iterate through ``self``.

        EXAMPLES::

            sage: M = SixVertexModel(2, boundary_conditions='ice')
            sage: list(M)
            [
                ^    ^          ^    ^
                |    |          |    |
            --> # <- # <--  --> # -> # <--
                |    ^          ^    |
                V    |          |    V
            --> # -> # <--  --> # <- # <--
                |    |          |    |
                V    V    ,     V    V
            ]
        """
        # Boundary conditions ordered URDL
        # The top row boundary condition of True is a downward arrow
        # The left condition of True is a right arrow
        # verts = ['LR', 'LU', 'LD', 'UD', 'UR', 'RD']
        next_top = [False, False, True, True, False, True]
        next_left = [True, False, False, False, True, True]
        check_top = [True, False, True, False, False, True]
        check_left = [False, False, False, True, True, True]

        bdry = [self._bdry_cond[0]]
        lbd = list(self._bdry_cond[3]) + [None] # Dummy
        left = [[lbd[0]]]
        cur = [[-1]]
        n = self._nrows
        m = self._ncols
        # [[3, 1], [5, 3]]
        # [[4, 3], [3, 2]]

        while cur:
            # If we're at the last row
            if len(cur) > n:
                cur.pop()
                left.pop()
                # Check if all our bottom boundary conditions are satisfied
                if all(x is not self._bdry_cond[2][i]
                       for i, x in enumerate(bdry[-1])):
                    yield self.element_class(self, tuple(tuple(x) for x in cur))
                bdry.pop()

            # Find the next row
            row = cur[-1]
            l = left[-1]
            i = len(cur) - 1
            while row:
                row[-1] += 1
                # Check to see if we have more vertices
                if row[-1] > 5:
                    row.pop()
                    l.pop()
                    continue
                # Check to see if we can add the vertex
                if (check_left[row[-1]] is l[-1] or l[-1] is None) \
                        and (check_top[row[-1]] is bdry[-1][len(row)-1]
                             or bdry[-1][len(row)-1] is None):
                    if len(row) != m:
                        l.append(next_left[row[-1]])
                        row.append(-1)
                    # Check the right bdry condition since we are at the rightmost entry
                    elif next_left[row[-1]] is not self._bdry_cond[1][i]:
                        bdry.append([next_top[x] for x in row])
                        cur.append([-1])
                        left.append([lbd[i+1]])
                        break

            # If we've killed this row, backup
            if len(row) == 0:
                cur.pop()
                bdry.pop()
                left.pop()

    def boundary_conditions(self):
        """
        Return the boundary conditions of ``self``.

        EXAMPLES::

            sage: M = SixVertexModel(2, boundary_conditions='ice')
            sage: M.boundary_conditions()
            ((False, False), (True, True), (False, False), (True, True))
        """
        return self._bdry_cond

    def partition_function(self, beta, epsilon):
        r"""
        Return the partition function of ``self``.

        The partition function of a 6 vertex model is defined by:

        .. MATH::

            Z = \sum_{\nu} e^{-\beta E(\nu)}

        where we sum over all configurations and `E` is the energy function.
        The constant `\beta` is known as the *inverse temperature* and is
        equal to `1 / k_B T` where `k_B` is Boltzmann's constant and `T` is
        the system's temperature.

        INPUT:

        - ``beta`` -- the inverse temperature constant `\beta`
        - ``epsilon`` -- the energy constants, see
          :meth:`~sage.combinat.six_vertex_model.SixVertexConfiguration.energy()`

        EXAMPLES::

            sage: M = SixVertexModel(3, boundary_conditions='ice')
            sage: M.partition_function(2, [1,2,1,2,1,2])                                # needs sage.symbolic
            e^(-24) + 2*e^(-28) + e^(-30) + 2*e^(-32) + e^(-36)

        REFERENCES:

        :wikipedia:`Partition_function_(statistical_mechanics)`
        """
        from sage.functions.log import exp
        return sum(exp(-beta * nu.energy(epsilon)) for nu in self)


class SquareIceModel(SixVertexModel):
    r"""
    The square ice model.

    The square ice model is a 6 vertex model on an `n \times n` grid with
    the boundary conditions that the top and bottom boundaries are pointing
    outward and the left and right boundaries are pointing inward. These
    boundary conditions are also called domain wall boundary conditions.

    Configurations of the 6 vertex model with domain wall boundary conditions
    are in bijection with alternating sign matrices.
    """

    def __init__(self, n):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: M = SixVertexModel(3, boundary_conditions='ice')
            sage: TestSuite(M).run()
        """
        boundary_conditions = ((False,)*n, (True,)*n)*2
        SixVertexModel.__init__(self, n, n, boundary_conditions)

    def from_alternating_sign_matrix(self, asm):
        """
        Return a configuration from the alternating sign matrix ``asm``.

        EXAMPLES::

            sage: M = SixVertexModel(3, boundary_conditions='ice')
            sage: asm = AlternatingSignMatrix([[0,1,0],[1,-1,1],[0,1,0]])               # needs sage.modules
            sage: M.from_alternating_sign_matrix(asm)                                   # needs sage.modules
                ^    ^    ^
                |    |    |
            --> # -> # <- # <--
                ^    |    ^
                |    V    |
            --> # <- # -> # <--
                |    ^    |
                V    |    V
            --> # -> # <- # <--
                |    |    |
                V    V    V

        TESTS::

            sage: M = SixVertexModel(5, boundary_conditions='ice')
            sage: ASM = AlternatingSignMatrices(5)                                      # needs sage.modules
            sage: all(M.from_alternating_sign_matrix(x.to_alternating_sign_matrix()) == x           # needs sage.modules
            ....:     for x in M)
            True
            sage: all(M.from_alternating_sign_matrix(x).to_alternating_sign_matrix() == x           # needs sage.modules
            ....:     for x in ASM)
            True
        """
        if asm.parent().size() != self._nrows:
            raise ValueError("mismatched size")

        #verts = ['LR', 'LU', 'LD', 'UD', 'UR', 'RD']
        ret = []
        bdry = [False]*self._nrows # False = up
        for row in asm.to_matrix():
            cur = []
            right = True # True = right
            for j,entry in enumerate(row):
                if entry == -1:
                    cur.append(0)
                    right = True
                    bdry[j] = False
                elif entry == 1:
                    cur.append(3)
                    right = False
                    bdry[j] = True
                else: # entry == 0
                    if bdry[j]:
                        if right:
                            cur.append(5)
                        else:
                            cur.append(2)
                    else:
                        if right:
                            cur.append(4)
                        else:
                            cur.append(1)
            ret.append(tuple(cur))
        return self.element_class(self, tuple(ret))

    class Element(SixVertexConfiguration):
        """
        An element in the square ice model.
        """
        @combinatorial_map(name='to alternating sign matrix')
        def to_alternating_sign_matrix(self):
            """
            Return an alternating sign matrix of ``self``.

            .. SEEALSO::

                :meth:`~sage.combinat.six_vertex_model.SixVertexConfiguration.to_signed_matrix()`

            EXAMPLES::

                sage: M = SixVertexModel(4, boundary_conditions='ice')
                sage: M[6].to_alternating_sign_matrix()                                 # needs sage.modules
                [1 0 0 0]
                [0 0 0 1]
                [0 0 1 0]
                [0 1 0 0]
                sage: M[7].to_alternating_sign_matrix()                                 # needs sage.modules
                [ 0  1  0  0]
                [ 1 -1  1  0]
                [ 0  1 -1  1]
                [ 0  0  1  0]
            """
            from sage.combinat.alternating_sign_matrix import AlternatingSignMatrix  # AlternatingSignMatrices
            # ASM = AlternatingSignMatrices(self.parent()._nrows)
            # return ASM(self.to_signed_matrix())
            return AlternatingSignMatrix(self.to_signed_matrix())



# sage_setup: distribution = sagemath-combinat

from sage.databases.all__sagemath_categories import *

from sage.misc.lazy_import import lazy_import

lazy_import('sage.databases.sloane', 'SloaneEncyclopedia')

lazy_import('sage.databases.oeis', 'oeis')

lazy_import('sage.databases.findstat', ['findstat', 'findmap'])

del lazy_import
